from __future__ import annotations  # https://stackoverflow.com/a/33533514

# system modules
import dataclasses
import math
import re
import json
import functools
import logging
from typing import Dict, Union, Literal, Sequence, Optional, Any, Iterator, Tuple


# external modules
from rich.text import Text


logger = logging.getLogger(__name__)


def make_it_n(items: Sequence, n: int, filler: Optional[Any] = None) -> Iterator:
    it = iter(items)
    for i in range(n):
        yield next(it, filler)


def make_it_two(items: Sequence, filler: Optional[Any] = None) -> Tuple[Any, Any]:
    it = iter(items)
    return next(it, filler), next(it, filler)


def sign(x: Union[float, int]) -> Union[Literal[-1, 1]]:
    return 1 if x >= 0 else -1


def from_jsonlines(string):
    if hasattr(string, "decode"):
        string = string.decode(errors="ignore")
    string = str(string or "")
    for i, line in enumerate(string.splitlines(), start=1):
        try:
            yield json.loads(line)
        except (json.JSONDecodeError, ValueError, TypeError) as e:
            logger.warning(f"line #{i} ({line!r}) is invalid JSON: {e!r}")
            continue


def as_regex(string: str, **kwargs):
    try:
        return re.compile(string, **kwargs)
    except Exception as e:
        logger.warning(f"Invalid regular rexpression {string!r}. Matching literally.")
        return re.compile(re.escape(string), **kwargs)


# TODO: How to type this correctly?
# Apparently there is no annotation for something 'Orderable' 🤔


@dataclasses.dataclass
class Range:
    start: object
    end: object

    def overlaps(self, other: Range) -> bool:
        return self.start <= other.end and other.start <= self.end  # type: ignore[operator]

    def extend(self, other: Range) -> Range:
        self.start = min(self.start, other.start)  # type: ignore[call-overload]
        self.end = max(self.end, other.end)  # type: ignore[call-overload]
        return self


@dataclasses.dataclass
class RangeMerger:
    ranges: list[Range] = dataclasses.field(default_factory=list)

    def add(self, newrg: Range):
        for rg in self.ranges:
            if rg.overlaps(newrg):
                rg.extend(newrg)
                return
        self.ranges.append(newrg)  # no overlap, start new range

    def collapse(self) -> RangeMerger:
        lastranges: list[Range] = []
        while (ranges := self.ranges.copy()) != lastranges:
            self.ranges.clear()
            for rg in ranges:
                self.add(rg)
            lastranges = ranges
        return self

    @property
    def total(self) -> object:
        self.collapse()
        if not self.ranges:
            return None
        return functools.reduce(
            lambda x, y: x + y, (rg.end - rg.start for rg in self.ranges)  # type: ignore[operator]
        )



# system modules
import time
from fnmatch import fnmatchcase
import warnings
from datetime import date
import unittest
import copy
import itertools
import glob
import collections
import uuid
import os
import json
import re
import textwrap
import sys
import shlex
import logging
import subprocess
import argparse
from collections import defaultdict
from typing import List, Dict, Set, Optional, Sequence, Tuple
from pathlib import Path
import importlib.metadata

# internal modules
from annextimelog.repo import Event, AnnextimelogRepo
from annextimelog.log import stdout, stderr, setup_logging, setup_locale
from annextimelog.run import run
from annextimelog.token import *
from annextimelog import utils
from annextimelog.utils import Range, RangeMerger
from annextimelog.datetime import datetime, datetime as dt, timedelta, timedelta as td

# external modules
import rich
from rich.console import Console
from rich.syntax import Syntax
from rich.panel import Panel
from rich.columns import Columns
from rich.align import Align
from rich.table import Table
from rich.pretty import Pretty
from rich.text import Text
from rich import box

logger = logging.getLogger(__name__)


def test_cmd_handler(args, other_args):
    loader = unittest.TestLoader()
    logger.debug(f"🧪 Importing test suite")
    import annextimelog.test

    logger.info(f"🚀 Running test suite")
    testsuite = loader.loadTestsFromModule(annextimelog.test)
    result = unittest.TextTestRunner(
        verbosity=args.test_verbosity, buffer=args.test_verbosity <= 2
    ).run(testsuite)
    logger.debug(f"{result = }")
    if result.wasSuccessful():
        logger.info(f"✅ Test suite completed successfully")
    else:
        logger.error(f"💥 There were problems during testing.")
        sys.exit(1)


def git_cmd_handler(args, other_args):
    result = args.repo.run_git(subprocess.Popen, other_args)
    result.wait()
    sys.exit(result.returncode)


def config_cmd_handler(
    args: argparse.Namespace, other_args: Optional[List[str]] = None
):
    git_config_args = [
        (
            re.sub(r"^(annextimelog\.)?", "annextimelog.", a)
            if a in AnnextimelogRepo.ANNEXTIMELOG_CONFIG_KEYS
            else a
        )
        for a in (other_args or [])
    ]
    result = args.repo.run_git(subprocess.Popen, ["config"] + git_config_args)
    result.wait()
    sys.exit(result.returncode)


def sync_cmd_handler(args, other_args):
    sys.exit(0 if args.repo.sync(other_args) else 1)


def track_cmd_handler(args, other_args):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    try:
        event = args.repo.Event.from_cli(args.metadata)
    except ValueError as e:
        logger.critical(f"Invalid metadata {shlex.join(args.metadata)!r}: {e}")
        sys.exit(1)
    if not (event.start or event.end):
        logger.debug(f"Event is fully open. Let's start it now.")
        event.start = dt.now()
    if logger.getEffectiveLevel() < logging.DEBUG:
        logger.debug(f"Event before saving:")
        stderr.print(event.to_rich())
    stderr.print(event.to_rich())
    if args.dry_run:
        logger.info(f"--dry-run was given, so this new event is not stored.")
        return
    event.store()


def summary_cmd_handler(
    args: argparse.Namespace, other_args: Optional[List[str]] = None
):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    query = getattr(
        args,
        "query",
        (
            default := (
                []
                if args.repo.config.get("annextimelog.listall", "false") == "true"
                else ["today"]
            )
        ),
    )
    kwargs = dict(is_condition=True)
    try:
        if not (tokens := Token.from_strings(query, config=args.repo.config, **kwargs)):
            logger.debug(f"No query tokens, using current day as constraint")
            tokens = Token.from_strings(default, config=args.repo.config, **kwargs)
    except ValueError as e:
        logger.critical(f"Invalid query {shlex.join(query)!r}: {e}")
        sys.exit(1)
    logger.debug(f"atl ls: {tokens = }")
    events = sorted(
        args.repo.find_events(tokens),
        key=lambda e: e.end or e.start or dt.now().astimezone(),
    )

    if getattr(args, "id_only", None):
        for event in events:
            stdout.out(event.id)
        return

    total_seconds: float = 0
    closed_events: List[Event] = []
    open_events: List[Event] = []
    touched_days: Set[date] = set()
    rangeMerger = utils.RangeMerger()
    for event in events:
        start = event.start or dt.now().astimezone()
        end = event.end or dt.now().astimezone()
        rangeMerger.add(Range(start, end))
        if start and end:
            total_seconds += abs((end - start).total_seconds())
            closed_events.append(event)
            startdate = day = start.astimezone().date()
            enddate = ((end or dt.now()) - td(microseconds=1)).date()
            while day <= enddate:
                touched_days.add(day)
                day += td(days=1)
        if not (start and end):
            open_events.append(event)
        event.output(args)
    if total := rangeMerger.total:
        nonoverlap_seconds = total.total_seconds()  # type: ignore[attr-defined]
    else:
        nonoverlap_seconds = 0
    logger.info(
        f"Found {len(events)} events spanning {len(rangeMerger.ranges)} disjunct time ranges with a total non-overlapping time of {td(seconds=nonoverlap_seconds).pretty_duration(units='h')} ({td(seconds=total_seconds).pretty_duration(units='h')} double-counting overlaps), touching {len(touched_days)} days."
    )


def edit_cmd_handler(
    args: argparse.Namespace,
    other_args: Optional[List[str]] = None,
    tokens: Optional[Sequence[Token]] = None,
):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    if tokens:
        logger.debug(
            f"Using tokens {Token.join(tokens)!r} instead of {shlex.join(args.query)!r} to modify events"
        )
    else:
        try:
            logger.debug(f"Interpreting query {shlex.join(args.query)!r}")
            tokens = [
                t for t in Token.from_strings(args.query, config=args.repo.config) if t
            ]
        except ValueError as e:
            if logger.getEffectiveLevel() <= logging.DEBUG:
                logger.critical(f"Invalid query {shlex.join(args.query)!r}: {e}")
            logger.critical(
                f"To edit events, the query must contain matching conditions and modification sequences, "
                f"Your query {shlex.join(args.query)!r} does not specify both (add some -vvvv for more info). "
                f"This usually means you need to add some 'if' before your conditions and/or 'set' before your modifications. "
                f"See the following examples:"
            )
            stderr.log(
                Syntax(
                    textwrap.dedent(
                        f"""
            atl edit note="My note" where id=asdf # set note for event 'asdf'
            atl edit @home if location= yesterday # set location to 'home' for all events yesterday that have none set
            atl edit this week set project=secret @work # set location to 'work' and project=secret for all events in this week
            """
                    ).strip(),
                    "bash",
                )
            )
            sys.exit(1)
    conditions: List[Token] = []
    actions: List[Token] = []
    for token in tokens:
        match token:
            case Token(is_condition=True):
                conditions.append(token)
            case Token(is_condition=False):
                actions.append(token)
            case _:
                logger.warning(f"Ignoring {token!r}")
    if not actions:
        logger.info(
            f"No actions to perform for events matching {Token.join(conditions)!r}. "
        )
        return
    if conditions:
        if args.allow_edit_all:
            logger.warning(
                f"Ignoring --all as you also gave {len(conditions)} conditions ({Token.join(conditions)!r})"
            )
    elif not args.allow_edit_all:
        logger.critical(
            f"No conditions given to update with {Token.join(actions)!r}. "
            f"If you really want to process all events, specify --all"
        )
        return
    logger.info(
        f"atl edit: Will modify events matching {Token.join(conditions)!r} with {Token.join(actions)!r}"
    )
    events = list(args.repo.find_events(conditions))
    changed_events: List[Event] = []
    for event in events:
        if logger.getEffectiveLevel() < logging.DEBUG:
            logger.debug(
                f"Event matching {Token.join(conditions)!r} before applying {Token.join(actions)!r}"
            )
            stderr.log(event.to_rich())
        changed_event = copy.deepcopy(event)
        changed_event.apply(actions)
        if event == changed_event:
            if logger.getEffectiveLevel() < logging.INFO or args.dry_run:
                logger.info(
                    f"Event {event.id} {event.title or 'untitled'!r} didn't change when applying {Token.join(actions)!r}"
                )
            continue
        if (
            args.repo.config.get("annextimelog.confirm", "true") == "true"
            or args.dry_run
            or logger.getEffectiveLevel() < logging.INFO
        ):
            logger.info(
                f"{'Unsaved changes' if args.dry_run else 'Changes'} to {event.id} "
                f"(which matches {Token.join(conditions)!r}) after applying {Token.join(actions)!r}:"
            )
            stderr.print(event.compare_to(changed_event))
        else:
            stderr.print(changed_event.to_rich())
        if args.dry_run:
            continue
        if changed_event.store():
            changed_events.append(changed_event)
        else:
            stderr.print()
    logger.info(f"{len(changed_events)} events were touched")


def stop_cmd_handler(args: argparse.Namespace, other_args: Optional[List[str]] = None):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    tokens = [t for t in Token.from_strings(args.query, reduce=False) if t]
    timetokens, other = Token.sort_into(tokens, TimeToken)  # type: ignore[type-abstract]
    logger.debug(f"{timetokens = }")
    logger.debug(f"{other = }")
    try:
        actions = list(Token.reduce(timetokens, is_condition=False))
    except ValueError as e:
        logger.critical(
            f"While interpreting {Token.join(timetokens)!r} as stopping time for open events matching {Token.join(other)!r}: {e}"
        )
        sys.exit(1)
    actions_: List[Token] = []
    match actions:
        case [Duration()]:
            actions_.extend(actions)
        case []:
            actions_.append(TimeEnd(time=dt.now()))
        case [Time(time=t) as tt]:
            actions_.append(TimeEnd(string=tt.string, time=t))
        case [TimeFrame(start=start, end=end) as tt]:
            actions_.append(TimeEnd(time=end or start or dt.now()))
        case _:
            logger.critical(
                f"Don't know how to interpret {Token.joinpretty(actions)} as end for open events matching {Token.join(other)!r}"
            )
            sys.exit(1)
    actions = [t.evolve(is_condition=False) for t in actions_]
    conditions = [t.evolve(is_condition=True) for t in other]
    logger.debug(f"{conditions = }")
    logger.debug(f"{actions = }")
    return edit_cmd_handler(
        args=args,
        tokens=[t for t in Token.from_strings(".openend", is_condition=True) if t]
        + conditions
        + actions,
    )


def continue_cmd_handler(
    args: argparse.Namespace, other_args: Optional[List[str]] = None
):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    tokens = [t for t in Token.from_strings(args.query, reduce=False) if t]
    actiontokens, othertokens = Token.split_by_trigger(tokens, ActionFollowingKeyword)
    timeactiontokens, conditions = Token.sort_into(othertokens, TimeToken)  # type: ignore[type-abstract]
    logger.debug(f"{timeactiontokens = }")
    logger.debug(f"{conditions = }")
    try:
        timeactions_ = list(Token.reduce(timeactiontokens, is_condition=False))
    except ValueError as e:
        logger.critical(
            f"While interpreting {Token.join(timeactiontokens)!r} as starting time for continuing events matching {Token.join(conditions)!r}: {e}"
        )
        sys.exit(1)
    timeactions: List[Token] = []
    now = dt.now()
    match timeactions_:
        case []:
            timeactions.append(TimeFrame(start=now))
        case [Time(time=t) as tt]:
            timeactions.append(TimeFrame(start=t))
        case [TimeFrame(start=start, end=end) as tt] if not start:
            timeactions.append(tt.evolve(start=now))
        case [TimeFrame(start=start, end=end) as tt]:
            timeactions.append(tt)
        case [Duration(duration as d)]:
            timeactions.append(TimeFrame(start=now, end=now + d))
        case _:
            logger.critical(
                f"Don't know how to interpret {Token.joinpretty(timeactions_)} "
                f"(from {Token.join(timeactiontokens)!r}) as start for continuing events matching {Token.join(conditions)!r}"
            )
            sys.exit(1)
    actions = [t.evolve(is_condition=False) for t in timeactions] + actiontokens
    conditions = [t.evolve(is_condition=True) for t in conditions]
    logger.debug(f"{conditions = }")
    logger.debug(f"{actions = }")
    matching_closed_events = list(
        args.repo.find_events(
            [t for t in Token.from_strings(".closed", is_condition=True) if t]
            + conditions
        )
    )
    logger.debug(
        f"Found {len(matching_closed_events)} closed events matching {Token.join(conditions)!r}"
    )
    past_events = [
        e
        for e in matching_closed_events
        if all(t <= dt.now().astimezone() for t in (e.start, e.end) if t)
    ]
    logger.debug(
        f"Of the {len(matching_closed_events)} closed events matching {Token.join(conditions)!r}, {len(past_events)} are in the past"
    )
    latest_events = sorted(
        past_events, key=lambda e: e.start or e.end or dt.now(), reverse=True
    )
    if event := next(iter(latest_events), None):
        newevent = event.evolve(id=None, paths=set())
        newevent.start = dt.now()
        newevent.end = None
        newevent.apply(actions)
    else:
        metadata = actions + conditions
        logger.info(
            f"Didn't find an closed past event matching {Token.join(conditions)!r}. Making a new one with {Token.join(metadata)!r} instead."
        )
        newevent = args.repo.Event()
        newevent.apply(metadata)
    stderr.print(newevent.to_rich())
    if args.dry_run:
        logger.info(f"--dry-run was given, so this new event is not stored.")
        return
    newevent.store()


def delete_cmd_handler(
    args: argparse.Namespace, other_args: Optional[List[str]] = None
):
    if other_args:
        logger.warning(f"🙈 Ignoring other arguments {other_args}")
    kwargs = dict(is_condition=True)
    try:
        tokens = Token.from_strings(args.query, config=args.repo.config, **kwargs)
    except ValueError as e:
        logger.critical(f"Invalid query {shlex.join(args.query)!r}: {e}")
        sys.exit(1)
    logger.debug(f"atl del: {tokens = }")
    events = list(args.repo.find_events(tokens))
    if not events:
        return
    logger.info(f"Will delete these {len(events)} events:")
    for event in events:
        stderr.print(event.to_rich())
    if args.dry_run:
        logger.info(f"--dry-run: not deleting the above events")
        return
    logger.info(f"Deleting the above {len(events)} events")
    args.repo.delete(events)


def key2value(x: str) -> Tuple[str, str]:
    if m := AnnextimelogRepo.GIT_CONFIG_REGEX.fullmatch(x):
        key, value = m.groups()
        if key not in AnnextimelogRepo.ANNEXTIMELOG_CONFIG_KEYS:
            logger.warning(f"{key!r} (from {x!r}) is not an annextimelog config key")
        if not value:
            value = "true"
        return key, value
    else:
        raise argparse.ArgumentTypeError(f"{x!r} is not a key=value pair")


def add_common_arguments(parser):
    # TODO return/yield new groups?
    datagroup = parser.add_argument_group(title="Data")
    datagroup.add_argument(
        "--repo",
        type=Path,
        default=(default := AnnextimelogRepo.DEFAULT_PATH),
        help=f"Backend repository to use. "
        f"Defaults to $ANNEXTIMELOGREPO, $ANNEXTIMELOG_REPO or $XDG_DATA_HOME/annextimelog (currently: {str(default)})",
    )
    datagroup.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="don't actually store, modify or delete events in the repo. "
        "Useful for testing what exactly commands would do."
        "Note that the automatic repo creation is still performed.",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Just do it. Ignore potential data loss.",
    )
    outputgroup = parser.add_argument_group(
        title="Output", description="Options changing output behaviour"
    )
    outputgroup.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="verbose output. More -v ⮕ more output",
    )
    outputgroup.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="less output. More -q ⮕ less output",
    )
    outputgroup.add_argument(
        "-O",
        "--output-format",
        choices=("rich", "console", "json", "timeclock", "cli"),
        default=(default := "console"),  # type: ignore
        help=f"Select output format. Defaults to {default!r}.",
    )


configdoc = "\n".join(
    f"> atl config {key} ... # {desc}"
    for key, desc in sorted(AnnextimelogRepo.ANNEXTIMELOG_CONFIG_KEYS.items())
)

parser = argparse.ArgumentParser(
    description="⏱️ Time tracker based on Git Annex",
    epilog=textwrap.dedent(
        f"""
🛠️ Usage

Logging events:

> atl tr work @office                # an open event starting now
> atl tr work @office since 10:10    # an open event starting at a specific time
> atl tr work @office 10:10 -        # same (note the dash)
> atl stop work           # some time later, sets end of the latest 'work' event
> atl cont work           # next day, copy the latest work event and start it now
> atl tr work for 4h @home with client=smallcorp on project=topsecret    # a past event of 4h duration
> atl tr 10 - 11 @doctor  # at the doctor's for 1h
> atl tr 10:00 title="Note for 10 o'clock"  # record a time point
> atl tr y22:00 - 30min ago sleep @home quality=meh
> atl -vvv tr ... # debug problems

    Note: Common prepositions like 'with', 'about', etc. are ignored. See the full list with
    > python -c 'from annextimelog.token import Noop;print(Noop.FILLERWORDS)'

Listing events:

> atl
> atl ls
> atl ls .open      # list events that don't have both start and end set
> atl ls .openend   # list events that don't have an end set („currently running”)
> atl ls week
> atl ls yesterday
> atl ls until 2days ago
> atl -O json ls -a  # dump all data as JSON
> atl -O timeclock ls -a | hledger -f timeclock:- bal --daily   # analyse with hledger

Editing events:

> atl stop                               # stop all open-end events
> atl mod .openend set end=now           # same but manually
> atl mod id=jajekwka project=secret     # add custom tag to specific event
> atl mod id=LXKGrBGA end=now            # change end of specific event
> atl mod id=LXKGrBGA until 18           # change end of specific event
> atl mod @home where note=text          # add 'home' to location where note contains 'text'

Continuing events:

> atl cont work @office           # continues the last matching closed event as a new event
> atl cont work @office until 15  # same, but sets the end time to 15:00 today
> atl cont work set project=A     # continue and add/override metadata

Removing events:

> atl rm id=O3YzvZ4m   # delete by ID
> atl rm @work sleep   # delete all events where you slept at work 😉

Syncing:

# add a git remote of your choice
> atl git remote add git@gitlab.com:you/yourrepo
# sync up
> atl sync

Configuration

> atl -c key=value ... # temporarily set config
> atl config key value # permanently set config
{configdoc}

    """.strip()
    ),
    prog="annextimelog",
    formatter_class=argparse.RawTextHelpFormatter,
)
parser.add_argument(
    "--no-config",
    action="store_true",
    help="Ignore config from git",
)
parser.add_argument(
    "-c",
    dest="extra_config",
    action="append",
    metavar="key=value",
    type=key2value,
    help="Set a temporary config 'key=value' or just 'key' (implicit '=true'). "
    "If not present, 'annextimelog.' will be prepended to the key. "
    f"The following keys are available: {', '.join(AnnextimelogRepo.ANNEXTIMELOG_CONFIG_KEYS)}.",
    default=[],
)
add_common_arguments(parser)
versiongroup = parser.add_mutually_exclusive_group()
versiongroup.add_argument(
    "--version",
    action="store_true",
    help="show version information and exit",
)
versiongroup.add_argument(
    "--version-only",
    action="store_true",
    help="show only version and exit",
)


subparsers = parser.add_subparsers(title="Subcommands", dest="subcommand")
testparser = subparsers.add_parser(
    "test",
    help="run test suite",
    description="Run the test suite",
    formatter_class=argparse.RawTextHelpFormatter,
)
testparser.add_argument(
    "-v",
    "--verbose",
    dest="test_verbosity",
    help="Increase verbosity of test runner. "
    "-v: show test names, "
    "-vv: show raw debug output in all tests, not just failed tests. "
    "(Note that to set the debug level of annextimelog itself, you need to specify 'atl -vvvvv text -vv') ",
    action="count",
    default=1,
)
testparser.set_defaults(func=test_cmd_handler)
gitparser = subparsers.add_parser(
    "git",
    help="Access the underlying git repository",
    add_help=False,
    formatter_class=argparse.RawTextHelpFormatter,
)
gitparser.set_defaults(func=git_cmd_handler)
configparser = subparsers.add_parser(
    "config",
    help="Convenience wrapper around 'atl git config [annextimelog.]key [value], "
    "e.g. 'atl config emojis false' will set the annextimelog.emojis config to false.",
    add_help=False,
    formatter_class=argparse.RawTextHelpFormatter,
)
configparser.set_defaults(func=config_cmd_handler)
syncparser = subparsers.add_parser(
    "sync",
    help="sync data",
    description=textwrap.dedent(
        """
    Sync data with configured remotes by running 'git annex assist'.
    """
    ).strip(),
    aliases=["sy"],
)
add_common_arguments(syncparser)
syncparser.set_defaults(func=sync_cmd_handler)
trackparser = subparsers.add_parser(
    "track",
    help="record a time period",
    description=textwrap.dedent(
        """
    Record a time with metadata.

    Example:

    > atl tr y22  800 work python @home ="annextimelog dev" :"working on cli etc." project=annextimelog project+=timetracker

    """
    ).strip(),
    aliases=["tr"],
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(trackparser)
trackparser.add_argument(
    "metadata",
    nargs="+",
    help=textwrap.dedent(
        """
    Examples:

        10:00                   10:00 today
        y15:00                  15:00 yesterday
        yy15:00                 15:00 the day before yesterday
        t20:00                  20:00 tomorrow
        tt20:00                 20:00 the day after tomorrow
        2023-12-04
        2023-12-04T13:15
        justaword               adds tag 'justaword'
        "with space"            (shell-quoted) adds tag "with space"
        field=value             sets metadata field 'field' to (only) 'value'
        field+=value            adds 'value' to metadata field
"""
    ).strip(),
)
trackparser.set_defaults(func=track_cmd_handler)
deleteparser = subparsers.add_parser(
    "delete",
    help="delete an event",
    description=textwrap.dedent(
        """
    Delete events matching a query similar to `atl ls`

    Example:

    # the following commands would delete event rejacisumuqedowe
    > atl del id=rejaci    # delete event with id containing string
    > atl del todo         # delete all events with a todo (tag or field)
    > atl delete yesterday # delete all events yesterday
    ...

    """
    ).strip(),
    aliases=["del", "rm", "remove"],
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(deleteparser)
deleteparser.add_argument("query", nargs="+")
deleteparser.set_defaults(func=delete_cmd_handler)
editparser = subparsers.add_parser(
    "edit",
    help="modify an event",
    description=textwrap.dedent(
        """
    Modify an event.
    """
    ).strip(),
    aliases=["ed", "mod", "change", "upd", "update"],
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(editparser)
editparser.add_argument("query", nargs="+")
editparser.add_argument(
    "-a",
    "--all",
    dest="allow_edit_all",
    help="Allow processing all events",
    action="store_true",
)
editparser.set_defaults(func=edit_cmd_handler)
stopparser = subparsers.add_parser(
    "stop",
    help="set end of seleted currently open-end events to now",
    description=textwrap.dedent(
        """
    Have selected currently open-end events stop now
    """
    ).strip(),
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(stopparser)
stopparser.add_argument("query", nargs="*")
stopparser.add_argument(
    "-a",
    "--all",
    dest="allow_edit_all",
    help="Allow processing all events",
    action="store_true",
)
stopparser.set_defaults(func=stop_cmd_handler)
contparser = subparsers.add_parser(
    "cont",
    help=(doc := "continue a closed event now or at a given time"),
    description=textwrap.dedent(
        f"""
    {doc}
    """
    ).strip(),
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(contparser)
contparser.add_argument("query", nargs="*")
contparser.set_defaults(func=continue_cmd_handler)
summaryparser = subparsers.add_parser(
    "summary",
    help="show a summary of tracked periods",
    description=textwrap.dedent(
        """
    List a summary of tracked periods

    The format matches the 'atl tr' syntax, e.g.:

    atl ls today # (the default)
    atl ls week
    atl ls month
    atl ls since 10min ago
    atl ls field=REGEX   # field has value matching a regular expression
    atl ls field=REGEX,REGEX,REGEX   # field has value matching any of the given regex
    atl ls @home last week
    atl ls .openend
    ...
    """
    ).strip(),
    aliases=["su", "ls", "list", "find", "search"],
    formatter_class=argparse.RawTextHelpFormatter,
)
add_common_arguments(summaryparser)
summaryparser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="list all events (unless another time period is given)",
)
summaryparser.add_argument("--match", choices="all any".split())
listgroup = summaryparser.add_mutually_exclusive_group()
listgroup.add_argument(
    "--id",
    "--id-only",
    dest="id_only",
    action="store_true",
    help="only print IDs of matched events",
)
listgroup.add_argument("-l", "--long", action="store_true", help="more details")
summaryparser.add_argument("query", nargs="*")
summaryparser.set_defaults(func=summary_cmd_handler)


def cli(cmdline: Sequence[str] = sys.argv[1:]):
    args, other_args = parser.parse_known_args(args=cmdline)

    setup_logging(level=logging.INFO - (args.verbose - args.quiet) * 5)
    setup_locale()

    if not sys.stdin.isatty():
        logger.warning(
            f"annextimelog's cli is not yet stable, be careful relying on it in scripts."
        )

    logger.debug(f"{args = }")
    logger.debug(f"{other_args = }")

    if args.version or args.version_only:
        version = importlib.metadata.version(package := "annextimelog")
        urls: Dict[str, str] = dict(
            utils.make_it_two(re.split(r"\s*,\s*", s, maxsplit=1))
            for s in (
                importlib.metadata.metadata("annextimelog").get_all("project-url") or []
            )
        )
        author = importlib.metadata.metadata("annextimelog")["Author"]
        logger.warning(
            f"The displayed version {version!r} does not (yet) reflect development commits made after the release, "
            f"if you installed {package} from {urls['Repository']}."
        )
        if args.version_only:
            stdout.out(version)
            sys.exit(0)
        stdout.print(
            textwrap.dedent(
                f"""
                {package} v[b]{version}[/b] - A cli time tracker based on Git Annex
                by [b]{author}[/b] ({urls['Author on Mastodon']})
                Source code: {urls['Repository']}
                Changelog: {urls['Changelog']}
            """
            ).strip("\r\n")
        )
        sys.exit(0)

    args.repo = AnnextimelogRepo(args.repo)

    if not args.repo.ensure_git():
        sys.exit(1)

    if not args.no_config:
        args.repo.read_config()

    # apply extra configs
    args.repo.config.update(
        {
            re.sub(r"^(annextimelog\.)?", "annextimelog.", k): v
            for k, v in args.extra_config
        }
    )
    if args.dry_run:
        args.repo.config["annextimelog.dryrun"] = "true"
    if getattr(args, "long", None):
        args.repo.config["annextimelog.longlist"] = "true"
    if getattr(args, "all", None):
        args.repo.config["annextimelog.listall"] = "true"
    if m := getattr(args, "match", None):
        args.repo.config["annextimelog.matchcondition"] = m

    if args.repo.config.get("annextimelog.color", "true") == "false":
        stdout.no_color = True
        stderr.no_color = True

    if not args.repo.ensure_git_annex():
        sys.exit(1)

    # ✅ at this point, args.repo is a git annex repository

    if args.repo.config.get("annextimelog.commit", "true") == "true":
        if args.subcommand not in ["git"]:
            if args.repo.dirty:
                logger.warning(
                    f"🐛 The repo {args.repo.path} is not clean. "
                    f"This should not happen. Committing away the following changes:"
                )
                result = args.repo.run_git(subprocess.Popen, ["status"])
                with stderr.status("Committing..."):
                    result = args.repo.run_git(subprocess.run, ["annex", "add"])
                    result = args.repo.run_git(subprocess.run, ["add", "-A"])
                    result = args.repo.run_git(
                        subprocess.run,
                        ["commit", "-m", "🧹 Leftover changes"],
                    )
                result = args.repo.run_git(subprocess.run, ["status", "--porcelain"])
                if not (result.returncode or result.stderr):
                    logger.info(f"✅ Repo is now clean")
                else:
                    logger.warning(f"Commiting leftover changes didn't work.")

    # handle the subcommand
    # (when a subcommand is specified, the 'func' default is set to a callback function)
    if not getattr(args, "func", None):
        # default to 'atl summary'
        args.func = summary_cmd_handler
    try:
        args.func(args, other_args)
    finally:
        if (
            args.subcommand not in ["git"]
            and args.repo.config.get("annextimelog.commit", "true") == "true"
        ):
            result = args.repo.run_git(subprocess.run, ["status", "--porcelain"])
            if result.returncode or result.stdout or result.stderr:
                logger.warning(
                    f"🐛 This command left the repo {args.repo.path} in an unclean state. "
                    f"This should not happen. Consider investigating. "
                    f"The next time you run any 'annextimelog' command, these changes will be committed."
                )
                result = args.repo.run_git(subprocess.Popen, ["status"])


if __name__ == "__main__":
    cli()



# system modules
import logging
import locale

# external modules
from rich.console import Console
from rich.logging import RichHandler

logger = logging.getLogger(__name__)

stdout = Console()
stderr = Console(stderr=True)


def setup_locale():
    try:
        # Allow locale for strftime etc.
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error as e:
        logger.debug(
            f"Couldn't set locale. "
            f"Timestamps etc. will be displayed in a generic format."
        )


def setup_logging(**kwargs):
    default_kwargs = dict(
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                console=stderr,  # log to stderr
                rich_tracebacks=True,
                show_path=kwargs.get("level", logging.INFO) < logging.DEBUG - 10,
            )
        ],
    )
    logging.basicConfig(**{**default_kwargs, **kwargs})



from __future__ import annotations  # https://stackoverflow.com/a/33533514

# system modules
from functools import cached_property
import math
import re
import json
import logging
from datetime import datetime as datetime_, timedelta as timedelta_  # type: ignore[assignment]
from typing import Dict, Union, Literal, Optional
from zoneinfo import ZoneInfo

# internal modules
from annextimelog.utils import sign

# external modules
from rich.text import Text

logger = logging.getLogger(__name__)


class timedelta(timedelta_):
    def pretty_duration(self, units: str = "wdhms") -> Text:
        seconds = self.total_seconds()
        parts: Dict[str, float] = dict()
        bins = dict(w=7 * 24 * 60 * 60, d=24 * 60 * 60, h=60 * 60, m=60, s=1)
        colors = dict(w="purple", d="green", h="blue", m="red", s="yellow")
        for unit, s in {k: v for k, v in bins.items() if k in units}.items():
            parts[unit] = math.floor(seconds / s)
            seconds %= s
        if seconds and unit != "s":
            parts[unit] += seconds / s
        text = Text()
        for u, n in parts.items():
            if n:
                text.append(re.sub(r"\.0$", "", f"{n:.1f}")).append(u, style=colors[u])
        if not str(text):
            text.append("0")
        return text

    @classmethod
    def ensure(cls, d) -> timedelta:
        match d:
            case _ if isinstance(d, cls):
                return d
            case _ if hasattr(d, "total_seconds"):
                return cls(seconds=d.total_seconds())
            case float() | int():
                return cls(seconds=d)
            case _:
                raise ValueError(f"Can't convert {d!r} to {cls.__name__}")

    @classmethod
    def tryargs(cls, **kwargs) -> Optional[timedelta]:
        try:
            return cls(**kwargs)
        except TypeError as e:
            return None


class datetime(datetime_):
    FIELDS = "year month day hour minute second microsecond".split()
    WEEKDAYS = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday".split()
    MONTHS = "January February March April May June July August September October November December".split()

    @classmethod
    def ensure(cls, d) -> datetime:
        return (
            d if isinstance(d, cls) else cls(**{f: getattr(d, f) for f in cls.FIELDS})
        )

    def this(self, unit: str, offset: int = 0, weekstartssunday=False) -> datetime:
        match unit:
            case "year":
                span = timedelta(days=367)
                result = self.this("month").replace(month=1)
                for i in range(abs(offset)):
                    result += sign(offset) * span + span / 2
                    result = result.this("month").replace(month=1)
            case "month":
                span = timedelta(days=32)
                result = self.this("day").replace(day=1)
                for i in range(abs(offset)):
                    result += sign(offset) * span + span / 2
                    result = result.this("day").replace(day=1)
            case "week":
                today = self.this("day")
                result = today - timedelta(days=today.weekday())
                if weekstartssunday:
                    result -= timedelta(days=1)
                result += offset * timedelta(days=7)
            case str() as s if s in self.FIELDS:
                kwargs: Dict[str, int] = {}
                for field in self.FIELDS[::-1]:
                    if field == unit:
                        break
                    kwargs[field] = 1 if field in {"day"} else 0
                result = self.replace(**kwargs)  # type: ignore[arg-type]
                span = timedelta(**{f"{unit}s": 1})
                if offset:
                    result += offset * span + span / 2
                    result = result.replace(**kwargs)  # type: ignore[arg-type]
            case str() as given if (
                nth_month := next(
                    (
                        i
                        for i, month in enumerate(self.MONTHS, start=1)
                        if month.lower().startswith(given.lower())
                    ),
                    None,
                )
            ) is not None:
                result = self.this("month")
                return result.replace(year=result.year + offset, month=nth_month)
            case str() as given if (
                nth_weekday := next(
                    (
                        i
                        for i, weekday in enumerate(self.WEEKDAYS)
                        if weekday.lower().startswith(given.lower())
                    ),
                    None,
                )
            ) is not None:
                result = self.this("day")
                result += timedelta(days=nth_weekday - self.weekday())
                if weekstartssunday and nth_weekday == 6:
                    result -= timedelta(days=7)
                result += offset * timedelta(days=7)
            case _:
                raise ValueError(f"{unit!r} is an invalid unit")
        return result

    def next(self, unit: str, offset=1, **kwargs) -> datetime:
        offset = abs(offset)
        kwargs["offset"] = offset
        return self.this(unit, **kwargs)

    def prev(self, unit: str, offset=-1, **kwargs) -> datetime:
        offset = -abs(offset)
        kwargs["offset"] = offset
        return self.this(unit, **kwargs)

    # alias for prev()
    def last(self, *args, **kwargs) -> datetime:
        return self.prev(*args, **kwargs)

    def timeformat(self, timezone=ZoneInfo("UTC")) -> str:
        return self.astimezone(timezone).strftime("%Y-%m-%dT%H:%M:%S%z")


# we must override the min and max constants, which are still from the base class otherwise
datetime.min = datetime.ensure(datetime_.min)
datetime.max = datetime.ensure(datetime_.max)



# system modules
import shlex
import logging
import subprocess
from pathlib import Path
from typing import Optional, Union, Sequence, Callable

# internal modules
from annextimelog.log import stderr

# external modules
from rich.table import Table
from rich.syntax import Syntax
from rich.panel import Panel
from rich.text import Text
from rich.highlighter import ReprHighlighter, ISO8601Highlighter
from rich import box

logger = logging.getLogger(__name__)


def show_process_result(
    result, output_lexer: str = "txt", show_title: Optional[bool] = None
):
    table = Table(expand=True, padding=0, box=box.SIMPLE)
    if result.stdout:
        table.add_column("📢 STDOUT", ratio=1, justify="center")
    if result.stderr:
        table.add_column("⚠️  STDERR", ratio=1, justify="center")
    if not (result.stdout or result.stderr):
        table.add_column("*no output*")
    cols = []
    if result.stdout:
        cols.append(
            Syntax(
                result.stdout.rstrip(),
                lexer=output_lexer,
                line_numbers=True,
                word_wrap=True,
            )
        )
    if result.stderr:
        cols.append(
            Syntax(
                result.stderr.rstrip(),
                lexer="txt",
                line_numbers=True,
                word_wrap=True,
            )
        )
    if cols:
        table.add_row(*cols)

    if show_title or (
        show_title is None and logger.getEffectiveLevel() < logging.DEBUG - 10
    ):
        titleparts = [
            getattr(result, "title", None),
            f"(↩️  [b]return code {result.returncode}[/b])",
        ]
        if logger.getEffectiveLevel() > logging.DEBUG:
            titleparts.append(
                f"[code]{shlex.join(result.args)}[/code]",
            )
        table.title = "\n".join(filter(None, titleparts))
    stderr.print(Panel(table))


def run(
    runner: Callable,
    cmdline: Sequence[Union[str, Path]],
    return_error: bool = True,
    title: Optional[str] = None,
    output_lexer: str = "txt",
    debug_on_error: bool = True,
    stderr_is_error: bool = True,
    **kwargs,
):
    """
    Run a given ``cmdline`` with a :mod:`subprocess` runner (e.g
    :any:`subprocess.check_output`) with passed command-line arguments.
    If ``return_error`` is ``True``, a raised
    :any:`subprocess.CalledProcessError` is caught and returned. If it's
    ``None``, just ``None`` is returned. Otherwise the exception is bubbled up.
    """
    cmdline = list(map(str, cmdline))
    if logger.getEffectiveLevel() < logging.DEBUG:
        lines = []
        if title:
            lines.append(f"# {title}")
        lines.append(
            f"""# 🚀 Executing (📋 you could copy-paste this) in 📁 {kwargs.get("cwd") or Path.cwd()}:"""
        )
        lines.append(shlex.join(map(str, cmdline)))
        stderr.print(
            Syntax(
                "\n".join(lines),
                "bash",
                line_numbers=False,
                indent_guides=False,
                word_wrap=True,
                padding=0,
            )
        )
    if runner is subprocess.run:
        kwargs.setdefault("capture_output", True)
        kwargs.setdefault("check", False)
    try:
        result = runner(
            cmdline, **{**dict(text=True, encoding="utf-8", errors="ignore"), **kwargs}
        )
    except subprocess.CalledProcessError as e:
        if return_error is True:
            return e
        elif return_error is None:
            return None
        else:
            raise
    result.title = title
    if (
        kwargs.get("capture_output") and logger.getEffectiveLevel() < logging.DEBUG - 5
    ) or (
        debug_on_error and (result.returncode or (result.stderr and stderr_is_error))
    ):
        if result.returncode or (result.stderr and stderr_is_error):
            logger.error(
                f"Something went wrong during {title or shlex.join(map(str,cmdline))}"
            )
        show_process_result(result, output_lexer=output_lexer, show_title=True)
    return result



from __future__ import annotations  # https://stackoverflow.com/a/33533514

# system modules
import os
import warnings
import re
import sys
import json
import time
from argparse import Namespace
import functools
import copy
import shlex
import subprocess
import locale
import logging
import textwrap
from collections import defaultdict
import string
import random
from datetime import date
from pathlib import Path
import dataclasses
from dataclasses import dataclass, asdict, fields, field
from typing import (
    cast,
    Any,
    Optional,
    Set,
    Dict,
    Iterable,
    Union,
    List,
    Tuple,
    DefaultDict,
    ClassVar,
    Mapping,
    Sequence,
    Literal,
)

# internal modules
from annextimelog.run import run
from annextimelog.log import stdout, stderr
from annextimelog.datetime import datetime, datetime as dt, timedelta
from annextimelog import utils
from annextimelog.token import *

# external modules
from rich.table import Table
from rich.align import Align
from rich.text import Text
from rich.console import Console, RenderableType
from rich.highlighter import ReprHighlighter, ISO8601Highlighter
from rich.prompt import Confirm
from rich import box

logger = logging.getLogger(__name__)


@dataclass
class AnnextimelogRepo:
    DEFAULT_PATH: ClassVar = Path(
        os.environ.get("ANNEXTIMELOGREPO")
        or os.environ.get("ANNEXTIMELOG_REPO")
        or Path(os.environ.get("XDG_DATA_HOME", Path.home() / ".local" / "share"))
        / "annextimelog"
    )
    GIT_CONFIG_REGEX: ClassVar = re.compile(
        r"^(?P<key>[^\s=]+)(?:=(?P<value>.*))?$", flags=re.IGNORECASE
    )
    ANNEXTIMELOG_CONFIG_KEYS: ClassVar = {
        "emojis": "whether emojis should be shown in pretty-formated event output",
        "color": "whether colored output should be used. The default 'true' decides based on the typical $TERM and $NO_COLOR envvars.",
        "commit": "whether events should be committed upon modification. "
        "Setting this to false can improve performance but will reduce granularity to undo changes. ",
        "confirm": "whether to ask before actually touching anything",
        "fast": "setting this to false will cause annextimelog be be more sloppy "
        "(and possible faster) by leaving out some non-critical cleanup steps. ",
        "weekstartssunday": "whether the week should start on Sunday instead of Monday (the default)",
        "longlist": "equivalent of specifying --long (e.g. atl ls -l)",
        "listall": "equivalent of specifying --all (e.g. atl ls -a)",
        "matchconditions": "equivalent of atl ls --match (set to 'all' or 'any')",
        "outputformat": "equivalent of -O / --output-format",
        "dryrun": "equivalent of -n / --dry-run",
        "slowdown": "a float value like 0.1 (ms) to make finding events *even* slower 😅 (for debugging purposes)",
    }
    OPEN_EVENTS_PATH: ClassVar = Path("open")

    path: Path = Path(DEFAULT_PATH)
    config: Dict[str, str] = field(default_factory=dict, compare=False, repr=False)

    def __getattr__(self, attr: str):
        @dataclass(repr=False)
        class RepoEvent(Event):  # Event class knowing it belongs to this repo
            repo: AnnextimelogRepo = field(default_factory=lambda: self)

        if attr.lower() in {"event"}:
            return RepoEvent
        raise AttributeError(f"{self.__class__.__name__} has no attribute {attr!r}")

    @property
    def dirty(self):
        result = self.run_git(
            subprocess.run,
            ["status", "--porcelain"],
            title="Checking if repo {self.path} is dirty",
        )
        return result.returncode or result.stdout or result.stderr

    @staticmethod
    def get_repo_root(path: Path) -> Union[Path, None]:
        logger.debug(f"🔎 Finding where the containing git repo root is for path {path}")
        result = run(
            subprocess.run,
            ["git", "-C", Path(path), "rev-parse", "--show-toplevel"],
            title=f"find git repo root for {path}",
        )
        if result.returncode:
            return None
        return Path(result.stdout.rstrip("\n").rstrip("\r"))

    @property
    def root(self) -> Union[Path, None]:
        return self.get_repo_root(self.path)

    @staticmethod
    def confirm(description: Optional[str] = None):
        def decorator(decorated_fun):
            @functools.wraps(decorated_fun)
            def wrapper(x, *args, **kwargs):
                config = x.config if isinstance(x, AnnextimelogRepo) else dict()
                argstr = f"{decorated_fun.__name__}({{}})".format(
                    ", ".join(
                        str(s)
                        for s in itertools.chain(
                            args, (f"{k}={v!r}" for k, v in kwargs.items())
                        )
                        if s
                    )
                )
                if config.get("annextimelog.confirm", "true") == "true":
                    if not sys.stdin.isatty():
                        logger.info(
                            f"Skipped {description or argstr!r} in non-interactive mode. "
                            f"Specify `atl -c confirm=false ...` to not ask for confirmation."
                        )
                        return
                    try:
                        answer = Confirm.ask(
                            description or f"Run {argstr}?", console=stdout
                        )
                    except KeyboardInterrupt:
                        return
                    except Exception as e:
                        logger.error(
                            f"Couldn't read answer for {description or argstr!r}: {e!r}"
                        )
                        return
                    if not answer:
                        logger.info(f"Skipped {description or argstr!r}")
                        return
                return decorated_fun(x, *args, **kwargs)

            return wrapper

        return decorator

    def run_git(self, runner, gitargs: Union[str | Sequence[str]], **kwargs):
        if isinstance(gitargs, str):
            gitargs = shlex.split(gitargs)
        return run(runner, ["git", "-C", str(self.path)] + list(gitargs), **kwargs)

    def ensure_git(self) -> bool:
        """
        Ensure this is a properly useable git repo
        """
        if self.path.exists() and not self.path.is_dir():
            logger.critical(f"{self.path} exists but is not a directory.")
            return False

        if self.path.exists():
            logger.debug(f"{self.path} exists")
            if repo_root := self.root:
                if repo_root.resolve() != self.path.resolve():
                    logger.critical(
                        f"There's something funny with {self.path}: git says the repo root is {repo_root}. "
                    )
                    return False
            else:
                logger.critical(f"{self.path} exists but is no git repository. 🤔")
                return False
        else:
            if not self.path.parent.exists():
                logger.info(f"📁 Creating {self.path.parent}")
                self.path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"Making a git repository at {self.path}")
            result = run(
                subprocess.run, ["git", "init", str(self.path)], capture_output=True
            )
            if result.returncode:
                logger.error(f"Couldn't make git repository at {self.path}")
                return False
        return True

    def ensure_git_annex(self) -> bool:
        """
        Ensure this is a git annex repo
        """
        logger.debug(f"Making sure {self.path} is a git annex repository")
        annex_uuid = None
        if not self.config:
            if not (
                result := self.run_git(
                    subprocess.run,
                    ["config", "annex.uuid"],
                    title=f"get (only) git annex repo uuid",
                )
            ).returncode:
                self.config["annex.uuid"] = result.stdout.strip()
        if not self.config.get("annex.uuid"):
            logger.debug(f"{self.path} is not a git annex repository")
            if not (
                result := self.run_git(
                    subprocess.run,
                    ["annex", "init"],
                    title=f"add an annex to {self.path}",
                )
            ).returncode:
                logger.info(f"Added an annex to {self.path}")
            else:
                logger.critical(f"Couldn't add an annex to {self.path}")
                return False
        return True

    def read_config(self):
        logger.debug(f"Reading config from repo {self.path}")
        result = self.run_git(subprocess.run, ["config", "--list"])
        for line in result.stdout.splitlines():
            if m := self.GIT_CONFIG_REGEX.fullmatch(line):
                self.config[m.group("key")] = m.group("value")
        if logger.getEffectiveLevel() < logging.DEBUG - 5:
            logger.debug(f"Read git config:\n{self.config}")
        if logger.getEffectiveLevel() < logging.DEBUG - 5:
            logger.debug(f"Config:\n{self.config}")

    def sync(self, args: Optional[Sequence[str]] = None) -> bool:
        """Sync with git annex"""
        if logger.getEffectiveLevel() < logging.DEBUG:
            with self.run_git(subprocess.Popen, ["annex", "assist"]) as process:
                process.wait()
                return process.returncode == 0
        else:
            with stderr.status("Syncing..."):
                result = self.run_git(
                    subprocess.run, ["annex", "assist"] + list(args or [])
                )
        if result.returncode or result.stderr:
            if result.returncode:
                logger.error(f"Syncing failed according to git annex.")
            if result.stderr:
                logger.warning(
                    f"git annex returned some STDERR messages. "
                    f"This might be harmless, maybe try again (a couple of times)."
                )
            return False
        else:
            logger.info(f"✅ Syncing finished")
        return True

    @confirm("Store event?")
    def store(self, event: Event) -> bool:
        """Store an event"""
        if is_new_event := not event.id:
            event.id = event.random_id()
        event.clean()
        # TODO: Better handling of this, see https://gitlab.com/nobodyinperson/annextimelog/-/issues/9
        if event.end and event.start and event.end < event.start:
            logger.info(
                f"↔️  event {event.id!r}: Swapping start and end (they're backwards)"
            )
            event.start, event.end = event.end, event.start

        folders: Set[Path] = set()
        if event.start and event.end:

            def event_in_folders():
                start, end = event.start, event.end
                month = start.this("month")
                while month <= end:
                    path = Path()
                    for p in "%Y %m".split():
                        path /= month.strftime(p)
                    yield path
                    month = month.next("month")

            for folder in event_in_folders():
                folders.add(folder)
        else:
            folders.add(self.OPEN_EVENTS_PATH)

        paths: Set[Path] = set()
        for folder in folders:
            if not (folder_ := self.path / folder).exists():
                logger.debug(f"📁 Creating new folder {folder}")
                folder_.mkdir(parents=True)
            file = (folder_ / event.id).with_suffix(event.SUFFIX)
            if (file.exists() or file.is_symlink()) and not (event.paths or event.key):
                logger.warning(
                    f"🐛 {file} exists although this event {event.id} is new (it has no paths or key attached). "
                    f"This is either a bug 🐛 or you just witnessed a collision. 💥"
                    f"🗑️ Removing {file}."
                )
                file.unlink()
            # TODO: It's a bit overkill to remove and remake symlinks, but it works
            if file.is_symlink() and not os.access(str(file), os.W_OK):
                logger.debug(f"🗑️ Removing existing read-only symlink {file}")
                file.unlink()
            file_existed = file.exists()
            with file.open("w") as fh:
                logger.debug(
                    f"🧾 {'Overwriting' if file_existed else 'Creating'} {file} with content {event.id!r}"
                )
                fh.write(event.id)
            try:
                paths.add(file.relative_to(self.path))
            except ValueError:
                paths.add(file)
        logger.debug(
            f"new paths for event {event.id}: {paths = }, current paths: {event.paths}"
        )
        if obsolete_paths := event.paths - paths:
            logger.debug(
                f"{len(obsolete_paths)} paths for event {event.id!r} are now obsolete:"
                f"\n{chr(10).join(map(str,obsolete_paths))}"
            )
            result = self.run_git(
                subprocess.run,
                ["rm", "-rf"] + sorted(obsolete_paths),  # type: ignore[arg-type]
            )
        logger.debug(f"event {event.id} now has paths {paths}")
        if any(p.stem == self.OPEN_EVENTS_PATH for p in paths):
            logger.critical(
                msg := f"THIS IS A BUG: Event {event.id} would have been saved to files {paths}, "
                f"which includes the open-events folder name {self.OPEN_EVENTS_PATH}. "
                f"This should not happen. "
            )
            raise Exception(msg)
        is_new = not event.paths  # if the event didn't have paths, it is considered new
        event.paths = paths
        with stderr.status(f"Adding {len(event.paths)} paths..."):
            result = self.run_git(
                subprocess.run,
                ["annex", "add", "--json"] + sorted(event.paths),  # type: ignore[arg-type]
                output_lexer="json",
                title=f"Adding {len(event.paths)} paths for event {event.id!r}",
            )
            keys = set()
            for info in utils.from_jsonlines(result.stdout):
                if key := info.get("key"):
                    keys.add(key)
            if len(keys) != 1:
                logger.warning(
                    f"🐛 Adding {len(event.paths)} paths for event {event.id!r} resulted in {len(keys)} keys {keys}. "
                    f"That should be exactly 1. This is probably a bug."
                )
            if keys:
                event.key = next(iter(keys), None)
                logger.debug(f"🔑 key for event {event.id!r} is {event.key!r}")

        assert event.key, "Key is not set, this should not happen"
        # TODO: do all the below git actions in parallel for speed
        if event.config.get("annextimelog.fast", "false") != "true":
            with stderr.status(f"Force-dropping {keys = }..."):
                result = self.run_git(
                    subprocess.run,
                    [
                        "annex",
                        "drop",
                        "--force",
                        "--key",
                    ]
                    + list(keys),
                    title=f"Force-dropping {keys = } for event {event.id!r}",
                )
        if self.dirty and event.config.get("annextimelog.commit", "true") == "true":
            with stderr.status(
                status := f"Committing {'addition' if is_new_event else 'movement'} of event {event.id!r}..."
            ):
                result = self.run_git(
                    subprocess.run,
                    [
                        "commit",
                        "-m",
                        (f"➕ Add {event.id}" if is_new_event else f"📝 Move {event.id}")
                        + (f" {event.title!r}" if event.title else ""),
                    ],
                    title=status,
                )
                if not result.returncode:
                    logger.info(
                        f"✅ Committed {'addition' if is_new_event else 'movement'} of event {event.id!r}"
                    )
        # TODO: Do this more intelligently: Find out which fields changed and
        # reflect that in the git annex command (or --batch input)
        # This current way is not very sync friendly (the most recent version
        # of the event wins, no merging)
        if not is_new:
            # reset metadata
            logger.debug(f"Removing all metadata for event {event.id}")
            logger.debug(
                f"Yup, this (full removal, then full back-addition) is how it's done currently "
                f"😅 It will probably be optimized at some point."
            )
            if self.run_git(
                subprocess.run,
                shlex.split("annex metadata --remove-all --key") + [event.key],
            ).returncode:
                logger.error(
                    f"Something went wrong purging annex metadata on event {event.id} key {event.key!r}"
                )
        # add metadata
        cmd = ["annex", "metadata", "--key", event.key]
        for field, values in event.fields.items():
            for value in values:
                if hasattr(value, "timeformat"):
                    value = value.timeformat()
                cmd.extend(["--set", f"{field}+={value}"])
        logger.debug(f"Setting all current metadata for event {event.id}")
        if self.run_git(subprocess.run, cmd).returncode:
            logger.error(
                f"Something went wrong setting annex metadata on event {event.id} key {event.key!r}"
            )
        if logger.getEffectiveLevel() <= logging.DEBUG and self.config.get(
            "annextimelog.outputformat"
        ) not in (
            "rich",
            "console",
        ):
            logger.debug(f"Event after saving:")
            stderr.print(
                event.to_rich(
                    long=self.config.get("annextimelog.longlist", "false") == "true"
                )
            )
        return True

    @confirm("Delete event?")
    def delete(self, events: Union[Event, Sequence[Event]]) -> bool:
        if isinstance(events, Event):
            events = [events]
        events = list(events)
        allpaths: Set[Path] = set()
        for event in events:
            allpaths.update(event.paths)
        if not allpaths:
            logger.info(f"No matching events to delete")
            return False
        with stderr.status(
            f"Deleting {len(allpaths)} paths for {len(events)} event{'s' if len(events)>1 else ''}"
        ) as status:
            result = self.run_git(
                subprocess.run, ["rm", "-rf"] + sorted(map(str, allpaths))
            )
            success = not (result.returncode or result.stderr)
            status.update(f"Committing deletion of {len(events)} events")
            if self.config.get("annextimelog.commit", "true") == "true":
                result = self.run_git(
                    subprocess.run,
                    [
                        "commit",
                        "-m",
                        f"🗑️ Remove event{'' if len(events) == 1 else 's'} {' '.join(map(str,(ev.id for ev in events)))}",
                    ],
                )
                success |= not (result.returncode or result.stderr)
        if success:
            logger.info(
                f"✅ Deleted {len(events)} {' '.join(sorted(map(str,(ev.id for ev in events))))}"
            )
        return success

    def find_events(
        self, query: Union[Sequence[Token] | Sequence[str] | str]
    ) -> Iterator[Event]:
        match query:
            case [*strings] if all(isinstance(s, str) for s in strings):
                query = Token.from_strings(strings, is_condition=True)  # type: ignore
            case str() as s:
                query = Token.from_strings(s, is_condition=True)  # type: ignore
        query = cast(Sequence[Token], query)  # make mypy happy
        if logger.getEffectiveLevel() < logging.DEBUG:
            logger.debug(f"Finding events matching these tokens: 👇")
            stderr.log(query)
        start = datetime.min
        end = datetime.max
        # keep intersection of all time frames as query frame
        for token in query:
            if not token.is_condition:
                logger.warning(f"Ignoring non-condition {token.string!r}")
                continue
            match token:
                case TimeFrame():
                    start = max(start, token.start or datetime.min)
                    end = min(end, token.end or datetime.max)
        logger.debug(f"{start = }, {end = }")
        # TODO: don't query all events but only paths that make sense
        # would be more elegant to use something like 'findkeys' which wouldn't output
        # duplicates, but then we'd have to use 'whereused' to find out the repo paths
        # and also 'findkeys' only lists existing non-missing annex keys, so meh...
        cmd = ["annex", "metadata", "--json"]
        cmd.extend(
            L := Event.git_annex_args_timerange(
                start=None if start == datetime.min else start,
                end=None if end == datetime.max else end,
            )
        )
        # if the searched events are open, only search the folder for open events
        if not (start and end) or any(
            (isinstance(t, Property) and getattr(t, "name", None) == "open")
            for t in query
        ):
            if (self.path / self.OPEN_EVENTS_PATH).is_dir():
                cmd.append(self.OPEN_EVENTS_PATH)
        logger.debug(f"git annex matching args: {L = }")
        with stderr.status(f"Querying metadata..."):
            result = self.run_git(subprocess.run, cmd)
        counter = itertools.count(counter_start := 1)
        events: List[Event] = []

        def statusmsg(n):
            return f"🔎 Searched {n} events, {len(events)} matched"

        try:
            slowdown = float(v := self.config.get("annextimelog.slowdown", "0"))
        except Exception as e:
            slowdown = 0.1
            logger.warning(
                f"Weird value {v}!r for slowdown config: {e!r}. Using {slowdown}."
            )
        with stderr.status(msg := statusmsg(0)) as status:
            for n, event in enumerate(
                self.Event.multiple_from_metadata(utils.from_jsonlines(result.stdout)),
                start=1,
            ):
                time.sleep(slowdown)
                event_matches = event.matches(
                    query, match=self.config.get("annextimelog.matchcondition", "all")
                )
                logger.debug(
                    f"Event {event.id} {'matches' if event_matches else 'does not match'} query {shlex.join(t.string for t in query)!r}"  # type: ignore
                )
                if event_matches:
                    events.append(event)
                    yield event
                status.update(msg := statusmsg(n))
        logger.debug(msg)


@dataclass(repr=False)
class Event:
    repo: Optional[AnnextimelogRepo] = None
    id: Optional[str] = None
    paths: Set[Path] = field(default_factory=set)
    key: Optional[str] = None
    fields: Dict[str, Set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )  # type: ignore

    SUFFIX: ClassVar = ".ev"
    RESERVED_FIELDS: ClassVar = "start end tag location title note id".split()

    @property
    def start(self) -> Union[datetime, None]:
        if "start" not in self.fields:
            self.fields["start"] = set()
        if not (start := self.fields["start"]):
            return None
        elif len(start) > 1:
            try:
                earliest = min(
                    d.astimezone() for d in (self.parse_date(s) for s in start) if d
                )
            except Exception as e:
                logger.error(
                    f"There are {len(start)} start times for event {self.id!r}, but I can't determine the earliest: {e!r}"
                )
                self.fields["start"].clear()
                return None
            logger.warning(
                f"There were {len(start)} start times for event {self.id!r}. Using the earlier one {earliest}."
            )
            self.fields["start"].clear()
            self.fields["start"].add(earliest.timeformat())
        return self.parse_date(next(iter(self.fields["start"]), None))

    @start.setter
    def start(self, value: Union[datetime, str, None]):
        self.fields["start"] = set()
        if value is None:
            return
        if d := self.parse_date(value):
            self.fields["start"].add(d.timeformat())
        else:
            logger.error(f"Couldn't interpret {value!r} as time.")

    @property
    def end(self) -> Union[datetime, None]:
        if "end" not in self.fields:
            self.fields["end"] = set()
        if not (end := self.fields["end"]):
            return None
        elif len(end) > 1:
            try:
                latest = max(
                    d.astimezone() for d in (self.parse_date(s) for s in end) if d
                )
            except Exception as e:
                logger.error(
                    f"There are {len(end)} end times for event {self.id!r}, but I can't determine the latest: {e!r}"
                )
                self.fields["end"].clear()
                return None
            logger.warning(
                f"There were {len(end)} end times for event {self.id!r}. Using the later one {latest}."
            )
            self.fields["end"].clear()
            self.fields["end"].add(latest.timeformat())
        return self.parse_date(next(iter(self.fields["end"]), None))

    @end.setter
    def end(self, value: Union[datetime, str, None]):
        self.fields["end"] = set()
        if value is None:
            return
        if d := self.parse_date(value):
            self.fields["end"].add(d.timeformat())
        else:
            logger.error(f"Couldn't interpret {value!r} as time.")

    @property
    def note(self):
        if len(note := self.fields.get("note", set())) > 1:
            note = "\n".join(self.fields["note"])
            self.fields["note"] = set()
            self.fields["note"].add(note)
        return "\n".join(self.fields.get("note", set()))

    @note.setter
    def note(self, value):
        self.fields["note"] = set([value])

    @property
    def title(self):
        if len(title := self.fields.get("title", set())) > 1 or any(
            re.search(r"[\r\n]", t) for t in title
        ):
            title = " ".join(re.sub(r"[\r\n]+", " ", t) for t in self.fields["title"])
            self.fields["title"] = set([title])
        return "\n".join(self.fields.get("title", set()))

    @title.setter
    def title(self, value):
        value = re.sub(r"[\r\n]+", " ", str(value))
        self.fields["title"] = set([value])

    @property
    def tags(self):
        if "tag" not in self.fields:
            self.fields["tag"] = set()
        return self.fields["tag"]

    @property
    def config(self) -> Dict[str, str]:
        return getattr(self.repo, "config", dict())

    @classmethod
    def classconfig(cls) -> Dict[str, str]:
        repo = None
        if repofield := next((f for f in fields(cls) if f.name == "repo"), None):
            try:
                repo = repofield.default_factory()  # type: ignore[misc]
            except Exception as e:
                repo = repofield.default
        return getattr(repo, "config", dict())

    @classmethod
    def multiple_from_metadata(
        cls, metadatalines: Sequence[Dict[str, Any]], **init_kwargs
    ) -> Iterator[Event]:
        keys: Dict[str, Dict[str, Any]] = defaultdict(lambda: defaultdict(set))
        for i, data in enumerate(metadatalines, start=1):
            if logger.getEffectiveLevel() < logging.DEBUG - 5:
                logger.debug(f"parsed git annex metadata line #{i}:\n{data}")
            if key := data.get("key"):
                keys[key]["data"] = data
            if p := data.get("file"):
                keys[key]["paths"].add(Path(p))
        for key, info in keys.items():
            if not (d := info.get("data")):
                continue
            event = cls.from_metadata(d, paths=info["paths"], **init_kwargs)
            if logger.getEffectiveLevel() < logging.DEBUG - 5:
                logger.debug(f"parsed Event from metadata line #{i}:\n{event}")
            yield event

    def evolve(self, **kwargs) -> Event:
        return dataclasses.replace(self, **kwargs)

    def clean(self):
        """
        Remove inconsistencies in this event.
        """
        properties = [
            attr
            for attr in dir(self)
            if isinstance(getattr(type(self), attr, None), property)
        ]
        # Call all properties - they do their own cleanup
        for p in properties:
            getattr(self, p)
        # remove empty fields
        for field in (empty_fields := [f for f, v in self.fields.items() if not v]):
            del self.fields[field]
        # ensure paths are Path
        self.paths = set(Path(p) for p in self.paths)

    @staticmethod
    def random_id():
        charsets = [
            random.choices(
                (
                    consonants := list(
                        set(string.ascii_lowercase) - set(vowels := "eaiou")
                    )
                ),
                k=(n := 8),
            ),
            random.choices(vowels, k=n),
        ]
        random.shuffle(charsets)
        return "".join(itertools.chain.from_iterable(zip(*charsets)))

    @staticmethod
    def parse_date(s) -> Union[datetime, None]:
        if isinstance(s, datetime):
            return s
        match Time.from_str(s):
            case Time(time=t):
                return t
            case TimeFrame(start=start, end=end):
                return start or end
        return None

    @classmethod
    def git_annex_args_timerange(cls, start=None, end=None):
        """
        Construct a git-annex matching expression suitable for use as arguments with :any:$(subprocess.run) to only match data files containing data in a given period of time based on the unix timestamp in the 'start' and 'end' metadata
        """
        data_starts_before_end_or_data_ends_after_start = shlex.split(
            "-( --metadata start<{end} --or --metadata end>{start} -)"
        )
        data_not_only_before_start = shlex.split(
            "--not -( --metadata start<{start} --and --metadata end<{start} -)"
        )
        data_not_only_after_end = shlex.split(
            "--not -( --metadata start>{end} --and --metadata end>{end} -)"
        )
        condition = []
        info = dict()
        start = cls.parse_date(start)
        end = cls.parse_date(end)
        if start is not None:
            condition += data_not_only_before_start
            info["start"] = start.timeformat()
        if end is not None:
            condition += data_not_only_after_end
            info["end"] = end.timeformat()
        if all(x is not None for x in (start, end)):
            condition += data_starts_before_end_or_data_ends_after_start
        return [p.format(**info) for p in condition]

    @staticmethod
    def timeformat(**kwargs) -> str:
        return datetime.timeformat(**kwargs)

    def apply(self, tokens: Sequence[Token]):
        unhandled: List[Token] = []
        now = dt.now().replace(microsecond=0)
        for token in tokens:
            match token:
                # we handle a Duration() here because I couln't figure out
                # how to robustly do it in Token.reduce()
                case Duration(duration=d) if (self.start and not self.end) or (
                    self.start and self.end and self.start == self.end
                ):
                    self.end = self.start + d
                case Duration(duration=d):
                    self.start, self.end = now - d, now
                case TimeStart(time=t):
                    self.start = t
                case TimeEnd(time=t):
                    self.end = t
                case TimeFrame(start=start, end=end):
                    self.start = start or self.start
                    self.end = end or self.end
                case FieldModifier(field="id"):
                    logger.warning(
                        f"Ignoring {token.string!r}: the id field is reserved"
                    )
                case SetField(field=field, values=values):
                    old = self.fields.get(field, set())
                    if old and old != values:
                        logger.debug(
                            f"Overwriting previous {field!r} values {old} with {values}"
                        )
                    self.fields[field] = set(values)
                case AddToField(field=field, values=values):
                    if field not in self.fields:
                        self.fields[field] = set()
                    for value in values:
                        self.fields[field].add(value)
                case RemoveFromField(field=field, values=values):
                    if field in self.fields:
                        for toremove in values:
                            regex = utils.as_regex(toremove)
                            for value in self.fields[field].copy():
                                if regex.search(value) and value in self.fields[field]:
                                    self.fields[field].remove(value)
                        if not self.fields[field]:
                            del self.fields[field]
                case UnsetField(field=field):
                    if field in self.fields:
                        del self.fields[field]
                case Noop():  # ignore noops, don't consider it unhandled
                    pass
                case _:
                    unhandled.append(token)
        match unhandled:
            case []:
                pass
            case [Time(time=t1), Time(time=t2)]:
                self.start, self.end = t1, t2
            case [Time(time=t)]:
                logger.warning(
                    f"{Token.join(unhandled)!r} ({'·'.join(t.__class__.__name__ for t in unhandled)}) is interpreted as timepoint. This behaviour is admittedly confusing and will probably be removed in a future version. To start the event at {t}, specify e.g. {Token.join([TimeKeywordSince('since')] + unhandled)!r}"
                )
                self.start, self.end = t, t
            case _:
                logger.warning(
                    f"Ignored {len(unhandled)} tokens {Token.join(unhandled)!r} ({'·'.join(t.__class__.__name__ for t in unhandled)})"
                )

    def matches(
        self,
        tokens: Union[Token, Sequence[Token]],
        match: Union[Literal["all", "any"]] = "all",
    ) -> Union[bool, None]:
        if isinstance(tokens, Token):  # only one token given
            token = tokens  # mypy doesn't see walrus-style here, so old school 🤷
            if token.is_condition is False:
                logger.warning(
                    f"{token} is not a condition token, still using it to match, but this shouldn't happend"
                )
            match token:
                case Time() | TimeStart() | TimeEnd() | TimeFrame():
                    # TODO: This below abomination of a code block is quite
                    # inelegant but necessary as the extremes (dt.min and dt.max)
                    # can't be safely astimezone()d or timestamp()ed (exceeds limits)
                    inputtimes = dict(
                        start=None,
                        end=None,
                        estart=self.start,
                        eend=self.end,
                    )
                    match token:
                        case Time(time=t):
                            inputtimes["start"] = t
                        case TimeFrame(start=t1, end=t2):
                            inputtimes["start"] = t1
                            inputtimes["end"] = t2
                        case _:
                            raise ValueError(
                                f"This shouldn't happen, weird time token {token!r}"
                            )

                    def to_float(k: str, v: Union[datetime, None]) -> float:
                        match v:
                            case datetime():
                                return v.timestamp()
                            case x if x in (dt.min, dt.ensure(dt.min)):
                                return float("-inf")
                            case x if x in (dt.max, dt.ensure(dt.max)):
                                return float("inf")
                            case None if "start" in k:
                                return float("-inf")
                            case None if "end" in k:
                                return float("inf")
                        raise ValueError(f"{k!r} = {v!r} is a weird time")

                    times = {k: to_float(k, v) for k, v in inputtimes.items()}
                    start = times["start"]
                    end = times["end"]
                    estart = times["estart"]
                    eend = times["eend"]
                    if logger.getEffectiveLevel() < logging.DEBUG - 5:
                        logger.debug(f"{inputtimes = }")
                        logger.debug(f"{times = }")
                    conditions = [
                        # event_starts_before_end_or_event_ends_after_start
                        estart <= end or eend >= start,
                        # event_not_only_before_start
                        not (estart <= start and end <= start),
                        # event_not_only_after_end
                        not (estart >= end and eend >= end),
                    ]
                    return all(conditions)
                # field contains given values
                case FieldValueModifier(field=field, values=patterns):
                    if (  # special case of plain 'tag' value given: match by field name
                        field in {"tag", "tags", None}
                        and "=" not in token.string
                        and len(patterns) == 1
                    ):
                        pattern = next(iter(patterns))
                        if [f for f in self.fields if f.lower() == pattern.lower()]:
                            return True
                    present_values = (
                        set([self.id or ""])  # id field special case
                        if field.lower() == "id"
                        else self.fields.get(field, set())
                    )
                    for pattern in patterns:
                        try:
                            regex = re.compile(pattern, flags=re.IGNORECASE)
                        except Exception as e:
                            logger.error(
                                f"{pattern!r} from {shlex.quote(token.string)} is not a valid regular expression: {e!r}. Skipping."
                            )
                            continue
                        matches = set(v for v in present_values if regex.search(v))
                        if logger.getEffectiveLevel() < logging.DEBUG - 5:
                            logger.debug(
                                f"{pattern = !r} matches {len(matches)} values {matches} in {field!r} field {present_values} of event {self.id}"
                            )
                        match token:
                            case RemoveFromField():
                                return not matches
                            case _ if not matches:
                                return False
                    return True
                # field is empty
                case UnsetField(field=field):
                    logger.debug(f"{self.fields.get(field,set()) = }")
                    return not self.fields.get(field, set())
                case Property(name="open"):
                    return not (self.start and self.end)
                case Property(name="closed"):
                    return (
                        bool(self.start)
                        and bool(self.end)
                        and not (self.start == self.end)
                    )
                case Property(name="hasend"):
                    return bool(self.end)
                case Property(name="hasstart"):
                    return bool(self.start)
                case Property(name="openend"):
                    return not self.end
                case Property(name="openstart"):
                    return not self.start
                case Property(name="timepoint"):
                    return bool((self.start and self.end) and self.start == self.end)
                case Property(name="empty"):
                    return not self.fields
                case Noop():
                    return True
                case _:
                    return None
        else:  # multiple tokens given
            handled: List[Token] = []
            unhandled: List[Token] = []
            results: List[bool] = []
            for token in tokens:
                match result := self.matches(token):
                    case None:
                        unhandled.append(token)
                    case _:
                        results.append(result)
                        handled.append(token)
            if unhandled:
                logger.warning(
                    f"Ignored {len(unhandled)} tokens {shlex.join(t.string for t in unhandled)!r} ({'·'.join(t.__class__.__name__ for t in unhandled)})"
                )
            if logger.getEffectiveLevel() < logging.DEBUG:
                for token, result in zip(handled, results):
                    logger.debug(
                        f"Event {self.id} {'matches' if result else 'does not match'} token {token!r}"
                    )
            match match:
                case "all":
                    return all(results)
                case "any":
                    return any(results)
                case _:
                    warnings.warn(f"weird {match = } argument. Using 'all'")
                    return all(results)
        return False

    @staticmethod
    def delegate_to_repo(decorated_fun):
        @functools.wraps(decorated_fun)
        def wrapper(self, *args, **kwargs):
            methodname = decorated_fun.__name__
            if not self.repo:
                raise ValueError(
                    f"Cannot {self.__class__.__name__}.{methodname}() without knowing what repo it belongs to."
                )
            if not (f := getattr(self.repo, methodname, None)):
                raise ValueError(f"{self.repo} has no method {methodname}")
            return f(self)

        return wrapper

    @delegate_to_repo
    def store(self):
        pass

    @delegate_to_repo
    def delete(self):
        pass

    #################
    ### 📥  Input ###
    #################
    @classmethod
    def from_metadata(cls, data: Dict[str, Any], **init_kwargs) -> Event:
        """
        Create an event from a parsed output line of ``git annex metadata --json``.
        """
        fields = data.get("fields", dict())
        kwargs = init_kwargs.copy()
        kwargs.setdefault("paths", set())
        if f := data.get("file"):
            kwargs["paths"].add(path := Path(f))
            kwargs["id"] = path.stem
        kwargs.update(
            dict(
                key=data.get("key"),
                fields={
                    k: set(v)
                    for k, v in fields.items()
                    if not (k.endswith("-lastchanged") or k in ["lastchanged"])
                },
            )
        )
        return cls(**kwargs)

    @classmethod
    def from_tokens(cls, tokens: Sequence[Token], **kwargs) -> Event:
        event = cls(**kwargs)
        event.apply(tokens)
        event.clean()
        return event

    @classmethod
    def from_cli(cls, cliargs: Sequence[str], **kwargs) -> Event:
        """
        Create a new event from command-line arguments such as given to 'atl track'
        """
        logger.debug(f"Creating event from {cliargs = }")
        for i, token in enumerate(
            tokens := Token.from_strings(
                cliargs, config=cls.classconfig(), is_condition=False
            ),
            start=1,
        ):
            logger.debug(f"arg #{i:2d}: {token!r}")
        return cls.from_tokens([t for t in tokens if t is not None], **kwargs)

    ##################
    ### 📢  Output ###
    ##################
    def to_rich(self, long=None) -> RenderableType:
        table = Table(
            title=self.title,
            padding=0,
            box=box.ROUNDED,
            show_header=False,
        )
        ImmutableFieldText = functools.partial(Text, style="bright_cyan")
        emojis = self.config.get("annextimelog.emojis", "true") == "true"
        if not any(self.to_dict().values()):
            table.add_column("")
            table.add_row(ImmutableFieldText("empty event"))
            return table
        table.add_column("Field", justify="right", style="cyan")
        table.add_column("Value", justify="left")
        longlist = (
            self.config.get("annextimelog.longlist", "false") == "true" or long is True
        )
        if self.id:
            table.add_row(ImmutableFieldText("id"), f"[b]{self.id}[/b]")

        def joinvalues(
            values: Iterable[str],
            sep: Optional[str] = None,
            bullet: Optional[str] = "·",
        ) -> str:
            if sep is None:
                sep = "\n" if any(" " in v for v in values) else " "
            if not emojis:
                bullet = "·"
            return sep.join(f"{bullet}{v}" for v in sorted(values))

        if self.paths and (longlist):
            table.add_row(
                "paths",
                ReprHighlighter()(Text(joinvalues([str(p) for p in self.paths]))),
            )
        if self.paths and longlist:
            table.add_row("key", self.key)
        timehighlighter = ISO8601Highlighter()
        start, end = self.start, self.end
        if start and end and start == end:
            table.add_row(
                ImmutableFieldText("time"), start.astimezone().strftime("%c%Z")
            )
            table.border_style = "blue"
        elif start or end:
            if start:
                table.add_row("start", start.astimezone().strftime("%c%Z"))
                table.border_style = "red"
            if end:
                table.add_row("end", end.astimezone().strftime("%c%Z"))
                table.border_style = "green"
            if start:
                text = td(
                    seconds=((end or dt.now()).astimezone() - start).total_seconds()
                ).pretty_duration()
                if not end:
                    if emojis:
                        # there's no Text.prepend() method...
                        text = Text("").join([Text("⏳ "), text])
                    text.append("...")
                table.add_row(ImmutableFieldText("duration"), text)
            if start and end:
                table.border_style = None
        if location := self.fields.get("location", set()):
            table.add_row("location", joinvalues(location, bullet="📍 "))
        if self.tags:
            table.add_row("tags", joinvalues(self.tags, bullet="🏷️ "))
        for field, values in sorted(self.fields.items()):
            if field in self.RESERVED_FIELDS:
                continue
            kw = dict()
            match field.lower():
                case "todo":
                    kw["bullet"] = "🔲 "
            table.add_row(field, joinvalues(values, **kw))
        if self.note:
            table.add_row("note", self.note)
        return table

    def compare_to(self, event: Event) -> RenderableType:
        """
        Make a pretty comparison between this event and another
        """
        table = Table.grid(expand=False)
        table.add_column("old", ratio=10)
        table.add_column(width=3, ratio=1)
        table.add_column("new", ratio=10)
        table.add_row(
            self.to_rich(),
            Align(Text("\n".join("→→→")), align="center", vertical="middle"),
            event.to_rich(),
        )
        return table

    def to_dict(self):
        if sys.version_info < (3, 12):
            # https://github.com/python/cpython/pull/32056
            # dataclasses.asdict() doesn't like defaultdict
            e = copy.copy(self)
            e.fields = dict(self.fields)  # turn defaultdict into plain dict
        else:
            e = self
        return asdict(
            e,
            dict_factory=lambda x: {
                k: getattr(v, "to_dict", lambda: v)() for k, v in x if k not in {"repo"}
            },
        )

    def to_json(self) -> str:
        def default(x):
            if hasattr(x, "timeformat"):
                return x.timeformat()
            if not isinstance(x, str):
                try:
                    iter(x)
                    return tuple(x)
                except TypeError:
                    pass
            return str(x)

        return json.dumps(self.to_dict(), default=default)

    def to_timeclock(self):
        def sanitize(s):
            s = re.sub(r"[,:;]", r"⁏", s)  # replace separation chars
            s = re.sub(r"[\r\n]+", r" ", s)  # no newlines
            return s

        hledger_tags = {
            k: " ⁏ ".join(map(sanitize, v))
            for k, v in self.fields.items()
            if k not in "start end".split()
        }
        for tag in sorted(self.tags):
            hledger_tags[tag] = ""
        hledger_tags = [f"{t}: {v}" for t, v in hledger_tags.items()]
        hledger_comment = f";  {', '.join(hledger_tags)}" if hledger_tags else ""
        info = [
            ":".join(self.fields.get("account", self.tags)) or "_",
            self.title,
            hledger_comment,
        ]
        return textwrap.dedent(
            f"""
        i {self.start.strftime('%Y-%m-%d %H:%M:%S%z')} {'  '.join(filter(bool,info))}
        o {self.end.strftime('%Y-%m-%d %H:%M:%S%z')}
        """
        ).strip()

    def to_cli(self) -> List[str]:
        args = []
        fields = self.fields.copy() if self.fields else dict()
        if start := self.start:
            fields.pop("start", None)
            args.append(start.timeformat(timezone=None))
        if end := self.end:
            fields.pop("end", None)
            args.append(end.timeformat(timezone=None))
        if tags := fields.pop("tag", None):
            args.extend(tags)
        for field, values in fields.items():
            for value in values:
                if hasattr(value, "timeformat"):
                    value = value.timeformat(timezone=None)
                args.append(f"{field}+={value}")
        return args

    def output(self, args: Namespace):
        printer = {
            "timeclock": print,
            "json": print,
            "cli": lambda args: print(shlex.join(["atl", "tr"] + self.to_cli())),
        }.get(args.output_format, stdout.print)
        printer(getattr(self, f"to_{args.output_format}", self.to_rich)())  # type: ignore[operator]

    def __repr__(self) -> str:
        if hasattr(sys, "ps1"):
            with (c := Console(force_terminal=False)).capture() as capture:
                c.print(self.to_rich(long=True))
            return capture.get()
        else:
            args = ", ".join(f"{f.name}={getattr(self,f.name)!r}" for f in fields(self))
            return f"{self.__class__.__name__}({args})"

    def __eq__(self, other):
        """
        Two events are considered equal if their fields match sensibly
        """

        def sanitize(fields):
            # ensure values are set
            fields = {
                k: (set([v]) if isinstance(v, str) else set(v))
                for k, v in fields.items()
                if v
            }
            # convert to common timezone
            fields = {
                field: {
                    (
                        v.astimezone(ZoneInfo("UTC")).replace(microsecond=0)
                        if hasattr(v, "astimezone")
                        else v
                    )
                    for v in values
                }
                for field, values in fields.items()
            }
            return fields

        fields = sanitize(self.fields)
        otherfields = sanitize(other.fields)
        for field, values in fields.items():
            othervalues = otherfields.pop(field, None)
            if not (values or othervalues):  # both empty
                continue
            if values != othervalues:
                if logger.getEffectiveLevel() < logging.DEBUG - 5:
                    logger.debug(
                        f"events aren't equal because {field!r} values {values} != {othervalues}"
                    )
                return False
        if any(otherfields.values()):
            if logger.getEffectiveLevel() < logging.DEBUG - 5:
                logger.debug(
                    f"events aren't equal there are values {set(otherfields.values)} left"
                )
            return False
        return True



from __future__ import annotations  # https://stackoverflow.com/a/33533514

# system modules
import re
import inspect
import itertools
import warnings
import functools
import shlex
import logging
from abc import ABC, abstractmethod
import dataclasses
from dataclasses import dataclass, field
from collections.abc import Sequence
from typing import (
    cast,
    Optional,
    Set,
    Dict,
    Union,
    List,
    Tuple,
    Iterator,
    ClassVar,
    FrozenSet,
)

# internal modules
from annextimelog.datetime import datetime, datetime as dt, timedelta, timedelta as td
from annextimelog.log import stderr


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Token(ABC):
    string: str = field(compare=False, default="", kw_only=True)
    is_condition: Optional[bool] = field(default=None, kw_only=True)

    @property
    def pretty(self) -> str:
        return str(self)

    @property
    def roundtrip(self) -> Union[Token | None]:
        for f in (
            lambda: type(self).from_str(str(self)),
            lambda: Token.from_str(str(self)),
        ):
            if t := f():
                return t
        return None

    @classmethod
    def recursive_subclasses(cls):
        yield cls
        for subcls in cls.__subclasses__():
            yield from subcls.recursive_subclasses()

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[Token, None]:
        """
        Recurse into subclasses and try their from_str() constructors.
        This exact method is inherited if it'string not overwritten (classmethods can't be abstractmethods...)
        """
        for subcls in cls.__subclasses__():
            if from_str := getattr(subcls, "from_str", None):
                try:
                    if token := from_str(string, **kwargs):
                        return token
                except Exception as e:
                    logger.error(
                        f"Calling {subcls.__name__}.from_str({string!r}) didn't work: {e.__class__.__name__} {e}"
                    )
        if cls is Token or str(cls) == str(Token):  # dirty hack: only in base class
            # fall back to just setting a tag
            return AddToField(
                string=string, field="tag", values=frozenset([string]), **kwargs
            )
        return None

    @staticmethod
    def join(tokens: Sequence[Token]) -> str:
        return shlex.join(t.string for t in tokens)

    @staticmethod
    def jointypes(tokens: Sequence[Token]) -> str:
        return "·".join(t.__class__.__name__ for t in tokens)

    @classmethod
    def joinpretty(cls, tokens: Sequence[Token]) -> str:
        return f"{cls.join(tokens)!r} [{cls.jointypes(tokens)}]"

    @classmethod
    def sort_into(
        cls, tokens: Sequence[Token], *categories: type[Token]
    ) -> List[List[Token]]:
        parts = [cast(List[Token], []) for c in categories]
        unhandled: List[Token] = []
        for token in tokens:
            nexttoken = False
            for i, category in enumerate(categories):
                if isinstance(token, category):
                    parts[i].append(token)
                    nexttoken = True
                    break
            if nexttoken:
                continue
            unhandled.append(token)
        return parts + [unhandled]

    @classmethod
    def split_by_trigger(
        cls,
        tokens: Sequence[Token],
        *triggers: type[Token],
        prohibited: Optional[Set[type[Token]]] = None,
        allow_repeat=False,
    ) -> List[List[Token]]:
        prohibited = prohibited or set()
        triggerparts = [cast(List[Token], []) for trigger in triggers]
        unhandled: List[Token] = []
        next_is = None
        found_triggers: Dict[int, Set[Token]] = dict()
        for token in tokens:
            if any(isinstance(token, t) for t in prohibited):
                raise ValueError(
                    f"in this context (splitting by {'/'.join(t.__name__ for t in triggers)}), "
                    f"{token.string!r} is prohibited"
                )
            is_trigger = False
            for i, trigger in enumerate(triggers):
                if isinstance(token, trigger):
                    already_had = ", ".join(
                        f"{t.string!r}" for t in found_triggers.get(i, set())
                    )
                    if next_is == i:
                        logger.info(
                            f"Superfluous {trigger.__name__} {token.string!r} "
                            f"(there was already {already_had})"
                        )
                    elif not allow_repeat and i in found_triggers:
                        raise ValueError(
                            f"There was already a {trigger.__name__} like {token.string!r} ({already_had}), repetition is not allowed"
                        )
                    next_is = i
                    is_trigger = True
                    if i not in found_triggers:
                        found_triggers[i] = set()
                    found_triggers[i].add(token)
                    break
            if is_trigger:  # don't include the triggers
                continue
            if next_is is None:
                unhandled.append(token)
            else:
                triggerparts[next_is].append(token)
        return triggerparts + [unhandled]

    @classmethod
    def reduce(
        cls,
        tokens: Sequence[Union[Token | None]],
        config: Optional[Dict[str, str]] = None,
        to: Optional[Sequence[type[Token]]] = None,
        **kwargs,
    ) -> Iterator[Token]:
        """
        Given a sequence of tokens, reduce it down by iteratively/recursively
        translate parts of it and dropping irrelevant elements.

        Args:
            tokens: the tokens
            config: the annextimelog git config dict
            **kwargs: further target attributes for Tokens, e.g. is_condition=...

        Steps:
            - sort into conditions and actions, recurse into those
            - sort into time-related tokens and the rest
            - recurse into time-related tokens to reduce into a single TimeFrame
        """
        config = config or dict()
        conf = dict(config=config)
        is_condition = kwargs.get("is_condition", None)
        cond = dict(is_condition=is_condition)
        now = dt.now().replace(microsecond=0)
        wss = dict(
            weekstartssunday=config.get(
                "annextimelog.weekstartssunday", "false"
            ).lower()
            == "true"
        )

        ### Preparation: consume all tokens into list and irrelevant ones ###
        tokens_: List[Token] = []
        for token in tokens:
            match token:
                case None | Noop():
                    logger.debug(f"Ignoring token {token!r}")
                case _:
                    tokens_.append(token)
        tokens = tokens_
        tokenstr = Token.join(tokens)
        ss = dict(string=(s := tokenstr), **kwargs)

        if not tokens:
            logger.debug(f"No tokens left, we're done.")
            yield from tokens
            return

        if logger.getEffectiveLevel() < logging.DEBUG:
            frame = inspect.currentframe()
            recursion_level = (
                len(
                    [f for f in inspect.getouterframes(frame) if f.function == "reduce"]
                )
                - 1
            )
            logger.debug(
                f"Attempting reduce({is_condition=},{to=}) ({recursion_level=}) of {len(tokens)} tokens {tokenstr!r} 👇"
            )
            stderr.log(tokens)

        # to=TokenClass given
        wanted: Sequence[type[Token]] = to or cast(Sequence[type[Token]], set())
        if wanted:
            # try to reduce the tokens
            result = list(cls.reduce(tokens, to=None, **cond, **conf))
            # check if it's reduces to exactly one matching token
            if len(result) == 1 and any(isinstance(result[0], c) for c in wanted):
                logger.debug(f"{tokenstr!r} does reduce down to {result[0]!r}")
                yield from result
            else:
                raise ValueError(
                    f"Couldn't reduce {tokenstr!r} into a single {' or '.join(t.__name__ for t in wanted)} "
                    f"(instead it reduces to {len(result)} tokens {Token.joinpretty(result)})"
                )
            return

        def reduce_fail_msg(problemtokens):
            return (
                f"Don't know how to interpret or reduce time-related tokens {Token.joinpretty(problemtokens)} (from {Token.join(tokens)!r}). "
                f"If you think this combination does makes sense, consider opening an issue (https://gitlab.com/nobodyinperson/annextimelog/-/issues/new) to discuss."
            )

        ### Step 1: Separate actions from conditions ###
        if None in (conds := set(t.is_condition for t in tokens)) or len(conds) > 1:
            next_is_condition: Optional[bool] = is_condition
            if next_is_condition is not None:
                logger.debug(
                    f"Because {is_condition=}, the following will be {'conditions' if is_condition else 'actions'}"
                )
            match is_condition:
                case None:
                    actions, conditions, unknown = Token.split_by_trigger(
                        tokens, ActionFollowingKeyword, ConditionFollowingKeyword
                    )
                case True:
                    conditions, unknown = Token.split_by_trigger(
                        tokens,
                        ConditionFollowingKeyword,
                        prohibited={ActionFollowingKeyword},
                    )
                    actions = []
                case False:
                    actions, unknown = Token.split_by_trigger(
                        tokens,
                        ActionFollowingKeyword,
                        prohibited={ConditionFollowingKeyword},
                    )
                    conditions = []
            if unknown:
                match bool(actions), bool(conditions):
                    case True, False:
                        logger.debug(
                            f"These previously unknown tokens {unknown} are conditions"
                        )
                        conditions.extend(unknown)
                    case False, True:
                        logger.debug(
                            f"These previously unknown tokens {unknown} are actions"
                        )
                        actions.extend(unknown)
                    case _ if is_condition is True:
                        logger.debug(
                            f"Assuming the unknown tokens {unknown} to be conditions, because {is_condition=}"
                        )
                        conditions.extend(unknown)
                    case _ if is_condition is False:
                        logger.debug(
                            f"Assuming the unknown tokens {unknown} to be actions, because {is_condition=}"
                        )
                        actions.extend(unknown)
                    case _ if any(
                        (
                            (
                                isinstance(t, FieldModifier)
                                and getattr(t, "field", "") == "id"
                            )
                            or isinstance(t, Property)
                        )
                        for t in unknown
                    ) and is_condition is None:
                        logger.info(
                            f"As it's unclear whether tokens {Token.join(unknown)!r} are actions or conditions in this context (no 'set', 'if', etc.), but there are tokens concerning the ID or are a property (starting with a dot), we use those as conditions and the rest as actions."
                        )
                        for token in unknown:
                            match token:
                                case FieldModifier(field="id") | Property():
                                    conditions.append(token)
                                case _:
                                    actions.append(token)

                    case _:
                        raise ValueError(
                            f"Don't know whether tokens {Token.joinpretty(unknown)} are actions or conditions in this context (no 'set', 'if', etc. and {is_condition=})."
                        )
            actions = [t.evolve(is_condition=False) for t in actions]
            conditions = [t.evolve(is_condition=True) for t in conditions]
            logger.debug(
                f"Found {len(conditions)} {conditions = } and {len(actions)} {actions = }"
            )
            if actions:
                logger.debug(
                    f"Reducing {len(actions)} actions {shlex.join(t.string for t in actions)!r}"
                )
                yield from cls.reduce(actions, config=config, is_condition=False)
            if conditions:
                logger.debug(
                    f"Reducing {len(conditions)} conditions {shlex.join(t.string for t in conditions)!r}"
                )
                yield from cls.reduce(conditions, config=config, is_condition=True)
            return

        ### no token is time-related, nothing more to do here
        if not any(isinstance(t, TimeToken) for t in tokens):
            logger.debug(
                f"None of the {len(tokens)} tokens {tokenstr!r} are time-related, so we're done!"
            )
            yield from tokens
            return

        ### all tokens are time-related, convert them to a TimeFrame
        if all(isinstance(t, TimeToken) for t in tokens):
            logger.debug(
                f"All {len(tokens)} tokens {tokenstr!r} are time-related, reducing them down"
            )
            sincetokens, untiltokens, unknowntokens = Token.split_by_trigger(
                tokens, TimeKeywordSince, TimeKeywordUntil
            )
            if logger.getEffectiveLevel() < logging.DEBUG:
                logger.debug(f"{sincetokens = }")
                logger.debug(f"{untiltokens = }")
                logger.debug(f"{unknowntokens = }")
            if unknowntokens and isinstance(tokens[-1], TimeKeywordUntil):
                logger.debug(
                    f"Last token of {Token.join(tokens)!r} is an 'until' (so open interval), meaning *since* {Token.join(unknowntokens)}"
                )
                sincetokens.extend(unknowntokens)
                unknowntokens.clear()
            if sincetokens or untiltokens:
                match bool(sincetokens), bool(untiltokens):
                    case True, False if unknowntokens:
                        logger.debug(
                            f"It's *since* {Token.join(sincetokens)!r}, so *until* {Token.join(unknowntokens)!r}"
                        )
                        untiltokens.extend(unknowntokens)
                    case False, True if unknowntokens:
                        logger.debug(
                            f"It's *until* {Token.join(untiltokens)!r}, so *since* {Token.join(unknowntokens)!r}"
                        )
                        sincetokens.extend(unknowntokens)
                framekwargs: Dict[str, Optional[datetime]] = dict(start=None, end=None)
                rangetokens: Dict[str, Optional[Token]] = dict(since=None, until=None)
                for key, tokens_ in dict(since=sincetokens, until=untiltokens).items():
                    rangetokens[key] = next(
                        cls.reduce(
                            tokens_,
                            to={Time, Duration, TimeFrame},  # type: ignore[arg-type]
                            **kwargs,
                        ),
                        None,
                    )
                    if not rangetokens[key]:
                        continue
                    match key, rangetokens[key]:
                        # don't handle Duration() here yet!
                        case "since", Time(time=t):
                            framekwargs["start"] = t
                        case ["since", Number() as nt] if (
                            tt := Time.from_str(nt.string)
                        ):
                            match tt:
                                case Time(time=t):
                                    framekwargs["end"] = t
                                case TimeFrame(start=t1, end=t2):
                                    framekwargs["end"] = t2 or t1
                        case "until", Time(time=t):
                            framekwargs["end"] = t
                        case ["until", Number() as nt] if (
                            tt := Time.from_str(nt.string)
                        ):
                            match tt:
                                case Time(time=t):
                                    framekwargs["end"] = t
                                case TimeFrame(start=t1, end=t2):
                                    framekwargs["end"] = t2 or t1
                        case "since", TimeFrame(start=t1, end=t2):
                            framekwargs["start"] = t1 or t2
                        case "until", TimeFrame(start=t1, end=t2):
                            framekwargs["end"] = t2 or t1
                logger.debug(f"{rangetokens = }")
                match rangetokens["since"], rangetokens["until"]:
                    case Duration(), Duration():
                        raise ValueError(f"What does {tokenstr!r} mean? 🤔")
                    case Duration(duration=d), _:
                        framekwargs["start"] = (framekwargs["end"] or now) - d
                    case _, Duration(duration=d):
                        framekwargs["end"] = (framekwargs["start"] or now) + d
                logger.debug(f"{framekwargs = }")
                yield TimeFrame(**framekwargs, **ss)  # type: ignore[arg-type]
                return

            logger.debug(
                f"None of the {len(tokens)} tokens {tokenstr!r} is associated to 'since' or 'until', so we're close to the bottom!"
            )

            # interpret the time-related tokens
            match unknowntokens:  # this is where the magic happens 🪄
                # these can't be simplified
                case [TimeFrame()] | [Time()] | [Duration()]:
                    yield from tokens
                case [Time(time=t1), Time(time=t2)]:
                    yield TimeFrame(start=t1, end=t2, **ss)
                case [TimeFrame(start=s1, end=e1), TimeFrame(start=s2, end=e2)]:
                    yield TimeFrame(start=s1 or e1, end=e2 or s1, **ss)
                # if it's a plain number, try to interpret it as a time like '2024' or '04.2023'
                case [Number() as number] if (t_ := Time.from_str(number.string)):
                    yield t_
                # 14 2h        15 for 2h
                case [Integer(value=n), Duration(duration=d)] if 0 <= n < 24:
                    t = dt.now().replace(hour=n, minute=0, second=0, microsecond=0)
                    yield TimeFrame(start=t, end=t + d, **ss)
                # 14:00 2h        15:14 for 3h
                case [Time(time=t), Duration(duration=d)]:
                    yield TimeFrame(start=t, end=t + d, **ss)
                # 10min2sec ago
                case [Duration(duration=d), TimeKeyword(name="ago")]:
                    yield Time(now - d, **ss)
                # long duration form:   1 hour   |    2 weeks   (but not '2 months' - unclear what that means)
                case [Integer(value=n), TimeKeywordUnit(name=unit) as ut] if (
                    d := td.tryargs(**{f"{unit}s": n})  # type: ignore[assignment]
                ):
                    yield Duration(duration=d, **ss)
                case [  # 10 min ago
                    Integer(value=n),
                    TimeKeywordUnit(name=unit) as ut,
                    TimeKeyword(name="ago"),
                ]:
                    match unit:
                        case "year" | "month" | "week":
                            Δt = now.this("day") - now.this(ut.name)
                            yield Time(now.this(unit, offset=-n) + Δt, **ss)
                        case _:
                            yield Time(now - td(**{f"{unit}s": n}), **ss)
                case [  # in 2h
                    TimeKeyword(name="in"),
                    Duration(duration=d),
                ]:
                    yield Time(now + d, **ss)
                case [  # in 4 weeks
                    TimeKeyword(name="in" | "after"),
                    Integer(value=n),
                    TimeKeywordUnit(name=unit) as ut,
                ]:
                    match unit:
                        case "year" | "month" | "week":
                            Δt = now.next("day") - now.this(ut.name)
                            yield Time(now.this(unit, offset=n) + Δt, **ss)
                        case _:
                            yield Time(now + td(**{f"{unit}s": n}), **ss)
                case [
                    Time(time=t),
                    Duration(duration=d),
                ]:  # 10:00 [for] 2h
                    yield TimeFrame(start=t, end=t + d, **ss)
                case [  # last 2h, next 10min
                    TimeKeywordIter() as w,
                    Duration(duration=d),
                ]:
                    yield TimeFrame(
                        start=now + w.n * d,
                        end=now + (w.n + 1) * d,
                        **ss,
                    )
                case [TimeKeywordUnit() as ut]:  # month, week, hour, etc...
                    yield TimeFrame(
                        start=now.this(ut.name, 0, **wss),
                        end=now.this(ut.name, 1, **wss),
                        **ss,
                    )
                case [  # last month, this week, next hour, etc.
                    TimeKeywordIter() as w,
                    TimeKeywordUnit() as ut,
                ]:
                    yield TimeFrame(
                        start=now.this(ut.name, w.n, **wss),
                        end=now.this(ut.name, w.n + 1, **wss),
                        **ss,
                    )
                # January  Feb  March nov dec ...
                case [TimeKeywordMonth() as mt]:
                    yield TimeFrame(start=mt.start(), end=mt.end(), **ss)
                # last december    next feb
                case [TimeKeywordIter() as it, TimeKeywordMonth() as mt]:
                    yield TimeFrame(
                        start=mt.start(offset=it.n), end=mt.end(offset=it.n), **ss
                    )
                case [  # this Monday
                    TimeKeywordIter() as i,
                    TimeKeywordDay() as day,
                ]:
                    yield TimeFrame(
                        start=now.this(day.name, offset=i.n, **wss),
                        end=now.this(day.name, offset=i.n, **wss) + td(days=1),
                        **ss,
                    )
                case [  # Friday 12[:00]
                    TimeKeywordDay() as day,
                    Time() | Number() as ttnt,
                    # a bit inefficient to re-convert the Time() here, but it reduces duplication
                ] if isinstance((tt := Time.from_str(ttnt.string)), Time):
                    yield Time(
                        time=now.this(day.name, **wss).replace(
                            hour=tt.time.hour,
                            minute=tt.time.minute,
                            second=tt.time.second,
                        ),
                        **ss,
                    )
                case [  # next Monday 10[:00]
                    TimeKeywordIter() as i,
                    TimeKeywordDay() as day,
                    Time() | Number() as ttnt,
                    # a bit inefficient to re-convert the Time() here, but it reduces duplication
                ] if isinstance((tt := Time.from_str(ttnt.string)), Time):
                    yield Time(
                        time=now.this(day.name, offset=i.n, **wss).replace(
                            hour=tt.time.hour,
                            minute=tt.time.minute,
                            second=tt.time.second,
                        ),
                        **ss,
                    )
                case [TimeKeywordDay() as day]:  # Monday Tueday Wed Fri ...
                    yield TimeFrame(start=day.start(**wss), end=day.end(**wss), **ss)
                case [
                    TimeKeywordDay() as day,
                    TimeKeywordIter() as i,
                    TimeKeywordUnit() as u,
                ]:  # Monday next week
                    p = now.this(u.name, offset=i.n)  # 'next week'
                    yield TimeFrame(
                        start=p.this(day.name),  # day in that week
                        end=p.this(day.name)
                        + td(days=1),  # end of that day / start of next day
                        **ss,
                    )
                case [TimeKeywordPeriod() as w]:
                    yield TimeFrame(start=w.start, end=w.end, **ss)
                case _:
                    logger.debug(msg := reduce_fail_msg(unknowntokens))
                    raise ValueError(msg)
            return
        else:  ### separation into time-related tokens and other tokens needed ###
            timetokens: List[TimeToken] = []
            othertokens: List[Token] = []
            # sort out the time-related tokens
            for token in tokens:
                match token:
                    case TimeToken():
                        if logger.getEffectiveLevel() < logging.DEBUG:
                            logger.debug(f"{token!r} is a TimeToken")
                        timetokens.append(token)
                    case _:
                        if logger.getEffectiveLevel() < logging.DEBUG:
                            logger.debug(f"{token!r} is NOT a TimeToken")
                        othertokens.append(token)
            logger.debug(
                f"{tokenstr!r} is split into {len(timetokens)} time-related tokens {timetokens} "
                f"and {len(othertokens)} other tokens {othertokens}"
            )
            if timetokens:
                yield from cls.reduce(timetokens, config=config, **kwargs)
            yield from othertokens

    @classmethod
    def from_strings(
        cls,
        strings: Union[str | Sequence[str]],
        reduce: bool = True,
        config: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Sequence[Union[Token | None]]:
        match strings:
            case str():
                strings = shlex.split(given := strings)
                logger.debug(f"from_strings({given!r}): split into {strings}")
        tokens = [cls.from_str(s) for s in strings]
        if reduce:
            tokens = list(cls.reduce(tokens, config=config, **kwargs))
        elif kwargs:
            warnings.warn(f"from_strings(): unused {kwargs = } with {reduce = }")
        return tokens

    def evolve(self, **kwargs) -> Token:
        return dataclasses.replace(self, **kwargs)

    @abstractmethod
    def __str__(self) -> str:
        pass

    @staticmethod
    def recurse_into_subclasses(decorated_fun):
        @classmethod
        @functools.wraps(decorated_fun)
        def wrapper(cls, *args, **kwargs):
            for subcls in cls.__subclasses__():
                if method := getattr(subcls, decorated_fun.__name__, None):
                    if token := method(*args, **kwargs):
                        return token
            return decorated_fun(cls, *args, **kwargs)

        return wrapper


@dataclass(frozen=True)
class Noop(Token):
    FILLERWORDS = set(
        sorted(
            """
        the about beside near to above between of towards across beyond off under after by
        on underneath against despite onto unlike along down opposite among
        during out up around except outside upon as over via past with at for before
        round within behind inside without below into than beneath like through
        """.split()
        )
    )

    @property
    def pretty(self) -> str:
        return f"{self.string!r}: do nothing"

    @Token.recurse_into_subclasses
    def from_str(cls, string: str, **kwargs) -> Union[Token, None]:
        if re.fullmatch(r"\s*", string) or string.lower() in cls.FILLERWORDS:
            return cls(string=string, **kwargs)  # type: ignore[operator]
        return None

    def __str__(self) -> str:
        return ""


@dataclass(frozen=True)
class ConditionCombinationKeyword(Noop):
    KEYWORDS = "and or not".split()

    @classmethod
    def from_str(
        cls, string: str, **kwargs
    ) -> Union[ConditionCombinationKeyword, None]:
        if string.lower() in cls.KEYWORDS:
            logger.warning(
                f"{string!r} does not actually alter how conditions are met or actions are performed (yet! Maybe it will in a future release...)"
            )
            return cls(string=string, **kwargs)
        return None

    def __str__(self) -> str:
        return self.string


@dataclass(frozen=True)
class ConditionFollowingKeyword(Token):
    KEYWORDS = "where while if but although despite when given".split()

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[ConditionFollowingKeyword, None]:
        if string.lower() in cls.KEYWORDS:
            return cls(string=string, **kwargs)
        return None

    def __str__(self) -> str:
        return self.string


@dataclass(frozen=True)
class ActionFollowingKeyword(Token):
    KEYWORDS = "do then set change modify mod edit update".split()

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[ActionFollowingKeyword, None]:
        if string.lower() in cls.KEYWORDS:
            return cls(string=string, **kwargs)
        return None

    def __str__(self) -> str:
        return self.string


@dataclass(frozen=True)
class TimeToken(Token):
    pass


@dataclass(frozen=True)
class Number(TimeToken):
    value: float


@dataclass(frozen=True)
class Integer(Number):
    value: int

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Optional[Integer]:
        try:
            return cls(value=int(string), string=string)
        except Exception:
            return None

    def __str__(self):
        return str(self.value)


@dataclass(frozen=True)
class Float(Number):
    value: float

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Optional[Float]:
        try:
            return cls(value=float(string), string=string)
        except Exception:
            return None

    def __str__(self):
        return str(self.value)


@dataclass(frozen=True)
class TimeKeyword(TimeToken):
    name: str
    KEYWORDS: ClassVar = set("ago in after".split())

    @Token.recurse_into_subclasses
    def from_str(cls, string: str, **kwargs) -> Union[TimeKeyword, None]:
        if string.lower() in cls.KEYWORDS:
            return cls(string=string, name=string, **kwargs)  # type: ignore[operator]
        return None

    def __str__(self) -> str:
        return self.name


@dataclass(frozen=True)
class TimeKeywordIter(TimeKeyword):
    KEYWORDS: ClassVar[Dict[str, int]] = {
        "this": 0,
        "next": 1,
        "coming": 1,
        "following": 1,
        "last": -1,
        "prev": -1,
        "previous": -1,
    }

    @property
    def n(self) -> int:
        return self.KEYWORDS.get(self.name, 0)


@dataclass(frozen=True)
class TimeKeywordUnit(TimeKeyword):
    KEYWORDS: ClassVar = tuple("second minute hour day week month year".split())

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[TimeKeywordUnit, None]:
        for kw in cls.KEYWORDS:
            if string.lower() in {kw, f"{kw}s"} or (
                kw in "second minute month"
                and len(string) >= 3
                and kw.startswith(string)
            ):
                return cls(string=string, name=kw, **kwargs)
        return None

    @property
    def one_smaller(self) -> TimeKeywordUnit:
        smallerunit = self.KEYWORDS[0]
        for smaller, unit in zip(self.KEYWORDS, list(self.KEYWORDS)[1:]):
            logger.debug(f"{smaller = }, {unit = }, {self.name = }")
            if unit == self.name:
                smallerunit = smaller
                break
        return cast(TimeKeywordUnit, self.evolve(name=smallerunit, string=smallerunit))


@dataclass(frozen=True)
class TimeKeywordPeriod(TimeKeyword):
    KEYWORDS: ClassVar = {"today": 0, "yesterday": -1, "tomorrow": 1}

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[TimeKeywordPeriod, None]:
        for kw in cls.KEYWORDS:
            if string.lower() == kw or (
                len(string) >= 3 and kw.startswith(string.lower())
            ):
                return cls(string=string, name=kw, **kwargs)
        return None

    @property
    def n(self) -> int:
        return self.KEYWORDS.get(self.name, 0)

    @property
    def start(self) -> datetime:
        return dt.now().this("day", offset=self.KEYWORDS.get(self.name, 0))

    @property
    def end(self) -> datetime:
        return dt.now().this("day", offset=self.KEYWORDS.get(self.name, 0) + 1)


@dataclass(frozen=True)
class TimeKeywordDay(TimeKeyword):
    # TODO: Can't use datetime.WEEKDAYS for some reason!?
    KEYWORDS: ClassVar = (
        "Monday Tuesday Wednesday Thursday Friday Saturday Sunday".split()
    )

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[TimeKeywordDay, None]:
        for kw in cls.KEYWORDS:
            if (
                string.lower() == kw.lower()  # monday tuesday wednesday ...
                # mon tues thursd fr ...
                or (len(string) >= 2 and kw.lower().startswith(string.lower()))
            ):
                return cls(string=string, name=kw, **kwargs)
        return None

    def start(self, **kwargs) -> datetime:
        return dt.now().this(self.name, **kwargs)

    def end(self, **kwargs) -> datetime:
        return self.start(**kwargs) + timedelta(days=1)


@dataclass(frozen=True)
class TimeKeywordMonth(TimeKeyword):
    KEYWORDS: ClassVar = "January February March April May June July August September October November December".split()

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Optional[TimeKeywordMonth]:
        for kw in cls.KEYWORDS:
            if (
                string.lower() == kw.lower()  # janua february march decemb...
                # jan feb mar...
                or (len(string) >= 3 and kw.lower().startswith(string.lower()))
            ):
                return cls(string=string, name=kw, **kwargs)
        return None

    def start(self, **kwargs) -> datetime:
        return dt.now().this(self.name, **kwargs)

    def end(self, **kwargs) -> datetime:
        return self.start(**kwargs).next("month")


@dataclass(frozen=True)
class TimeKeywordUntil(TimeKeyword):
    KEYWORDS = set("until til till to -".split())


@dataclass(frozen=True)
class TimeKeywordSince(TimeKeyword):
    KEYWORDS = set("since starting from".split())


@dataclass(frozen=True)
class Time(TimeToken):
    """
    A specification of a point in time, such as:

        10     # 10:00 today
        y10    # yesterday 10:00
        yy10   # day before yesterday 10:00
        t10    # tomorrow 10:00
        tt10   # day after tomorrow 10:00
        2023-12-30T13:13:40+0200    # full ISO format
        13:13:40 # partial full ISO format
        ...
    """

    time: datetime

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Optional[Union[Time, TimeFrame]]:
        if string is None:
            return None
        offset = timedelta(days=0)
        if m := re.search(r"^(?P<prefix>[yt]+)(?P<rest>.*)$", string):
            offset = timedelta(
                days=sum(dict(y=-1, t=1).get(c, 0) for c in m.group("prefix"))
            )
            if string := m.group("rest"):
                pass
                # logger.debug(
                #     f"{string!r} starts with {m.group('prefix')!r}, so thats as an {offset = }"
                # )
            else:
                logger.debug(f"{string!r} means an {offset = } from today")
                return cls(
                    string=string,
                    time=dt.now().replace(hour=0, minute=0, second=0, microsecond=0)
                    + offset,
                    **kwargs,
                )

        # Try the bigger periods
        for period, formats in dict(
            year="%Y".split(),
            month="%Y-%m    %Y/%m   %m.%Y   %Y%m".split(),
            day="%Y-%m-%d   %Y/%m/%d   %d.%m.%Y".split(),
        ).items():
            for fmt in formats:
                try:
                    r = dt.strptime(string, fmt)
                except Exception as e:
                    continue
                if not r:
                    continue
                return TimeFrame(
                    string=string, start=r.this(period), end=r.next(period)
                )

        # Try full time specifications
        if re.fullmatch(r"\d{3}", string):
            # prepend zero to '100', otherwise interpreted as 10:00
            string = f"0{string}"
        result = None
        todaystr = dt.now().strftime(todayfmt := "%Y-%m-%d")
        for i, f in enumerate(
            (
                lambda s: dt.now() if s == "now" else None,
                dt.fromisoformat,
                # lambda s: dt.fromisoformat(f"{todaystr} {s}"),
                # Python<3.11 fromisoformat is limited, we implement the basic formats here
                # so we need to do it manually...
                lambda s: dt.strptime(s, "%Y-%m-%dT%H:%M%z"),
                lambda s: dt.strptime(s, "%Y-%m-%d %H:%M%z"),
                lambda s: dt.strptime(s, "%Y-%m-%dT%H:%M:%S"),
                lambda s: dt.strptime(s, "%Y-%m-%d %H:%M:%S"),
                lambda s: dt.strptime(s, "%Y-%m-%dT%H:%M:%S%z"),
                lambda s: dt.strptime(s, "%Y-%m-%d %H:%M:%S%z"),
                # conflicts with just year, e.g. '2024'
                # lambda s: dt.strptime(f"{todaystr} {s}", f"{todayfmt} %H%M"),
                lambda s: dt.strptime(f"{todaystr} {s}", f"{todayfmt} %H"),
                lambda s: dt.strptime(f"{todaystr} {s}", f"{todayfmt} %H:%M"),
                lambda s: dt.strptime(s, "%Y-%m-%d %H%M"),
            )
        ):
            try:
                if result := f(string):
                    break
            except Exception as e:
                pass
        if result:
            result += offset
            return Time(string=string, time=result)

        return None

    def __str__(self) -> str:
        return self.time.strftime("%Y-%m-%dT%H:%M:%S%z")


@dataclass(frozen=True)
class TimeStart(Time):
    def __str__(self):
        return f"start={self.time.isoformat()}"


@dataclass(frozen=True)
class TimeEnd(Time):
    def __str__(self):
        return f"end={self.time.isoformat()}"


@dataclass(frozen=True)
class TimeFrame(TimeToken):
    start: Optional[datetime] = None
    end: Optional[datetime] = None

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[TimeFrame, None]:
        if m := re.fullmatch(r"[({\[](?P<start>.*);(?P<end>.*)[)}\]]", string.strip()):
            if not (start := Time.from_str(s := m.group("start"))) and s:
                return None
            if not (end := Time.from_str(s := m.group("end"))) and s:
                return None
            return cls(
                string=string,
                start=getattr(start, "time", None),
                end=getattr(end, "time", None),
                **kwargs,
            )
        return None

    def __str__(self):
        return f"[{';'.join((t.isoformat() if t else '') for t in (self.start, self.end))}]"


@dataclass(frozen=True)
class Duration(TimeToken):
    """
    A duration specified in the following format:

        10min
        2h30m
        1week2days3hours4minutes5seconds
        1w2d3h4m5s
        ...
    """

    duration: timedelta

    UNITS = ["weeks", "days", "hours", "minutes", "seconds"]

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[Duration, None]:
        durations: List[timedelta] = []
        matches: int = 0
        s = string
        while s:
            if m := re.match(
                rf"[^\da-z]*(?P<number>\d+(?:\.\d+)?)[^\da-z]*(?P<unit>[a-z]+)[^\da-z]*",
                s,
                flags=re.IGNORECASE,
            ):
                number, unit = m.groups()
                if kwarg := next(
                    (u for u in cls.UNITS if u.startswith(unit.lower())), None
                ):
                    durations.append(timedelta(**{kwarg: float(number)}))
                s = s[m.span()[-1] :]  # drop this match and go on
                continue
            else:
                return None
        if not durations:
            return None
        return cls(string=string, duration=sum(durations, start=timedelta(0)), **kwargs)

    def __str__(self) -> str:
        parts: List[Tuple[int, str]] = []
        duration = self.duration
        for unit in sorted(self.UNITS, key=lambda u: timedelta(**{u: 1}), reverse=True):
            unitdelta = timedelta(**{unit: 1})
            if abs(duration) < abs(unitdelta):
                continue
            unitblocks = duration // unitdelta
            parts.append((unitblocks, unit))
            duration -= unitblocks * unitdelta  # type: ignore[assignment]
        return "".join(f"{n}{u[0]}" for n, u in parts if n)


@dataclass(frozen=True)
class Property(Token):
    KEYWORDS: ClassVar = set(
        "open openstart openend hasend hasstart closed timepoint empty".split()
    )

    name: str

    @classmethod
    def from_str(cls, string: str, **kwargs) -> Union[Property, None]:
        for kw in cls.KEYWORDS:
            if f".{kw}" == string.lower():
                return cls(string=string, name=kw, **kwargs)
        return None

    def __str__(self) -> str:
        return f".{self.name}"


@dataclass(frozen=True)
class FieldModifier(Token):
    """
    A metadata field modifier such as:

        field=value          # set 'field' to (only) 'value'
        field+=value         # add 'value' to 'field'
        field=bla,bli,blubb  # set 'field' to given three values
        field+=bla,bli,blubb # add multiple values to 'field'
        field-=value         # remove 'value' from 'field'
        field-=bla,bli,blubb # remove multiple values from 'field'
        field+/=a,b,c/d,e,f  # different separator (this adds 'a,b,c' and 'd,e,f' to 'field')
    """

    field: str

    # don't want to put too many in here, syntax might be needed later
    SEPARATORS: ClassVar = ",;:"
    # we don't allow a dot in front (reserved for Property() token), otherwise
    # this is what git annex supports
    FIELD_NAME_REGEX: ClassVar = re.compile(r"^(?!\.)[\w_.-]+$", re.IGNORECASE)

    @classmethod
    def from_str(
        cls, string: str, **kwargs
    ) -> Union[FieldModifier, TimeStart, TimeEnd, Noop, None]:
        # short form
        if m := re.search(r"^(?P<symbol>[@:=])(?P<value>.*)$", string.strip()):
            field = {"@": "location", ":": "note", "=": "title"}.get(
                m.group("symbol"), ""
            )
            kwargs = dict(
                string=string, field=field, values=set([m.group("value")]), **kwargs
            )
            return SetField(**kwargs)
        # long form
        if m := re.search(
            rf"(?P<field>\S+?)(?P<operator>[+-]?)(?P<sep>[{cls.SEPARATORS}]?)=(?P<values>.*)",
            string,
        ):
            field, operator, sep, givenvalues = m.groups()
            if not cls.FIELD_NAME_REGEX.fullmatch(field):
                logger.warning(f"{field!r} is an invalid metadata field name")
                return None
            if not sep and re.search(r"\s+", givenvalues):
                values = [givenvalues.strip()]
            else:
                sep = sep or ","
                values = [
                    v.strip()
                    for v in re.split(rf"(?:{re.escape(sep)})+", givenvalues)
                    if v
                ]
            values = cast(List[str], values)
            match field.lower():
                case "start" | "end" as field_:
                    # start='in 10min', separated by whitespace, not comma like other fields
                    values = shlex.split(m.group("values"))
                    if operator in "-+".split():
                        logger.warning(
                            f"Ignoring {operator = !r} in {string} (start and end field are special)"
                        )
                        operator = ""
                    if not values:
                        return SetField(field_, frozenset(), string=string)
                    try:
                        timetokens: List[TimeToken] = []
                        othertokens: List[Token] = []
                        logger.debug(
                            f"{string!r}: Determining what {field_} time {shlex.join(values)!r} means"
                        )
                        for token in Token.from_strings(
                            list(values), is_condition=False
                        ):
                            match token:
                                case TimeToken():
                                    timetokens.append(token)
                                case t if t:
                                    othertokens.append(token)
                                    logger.warning(
                                        f"While parsing {values} (from {string!r}): Ignoring {token.string!r} as it is not time-related"
                                    )
                    except ValueError as e:
                        raise ValueError(
                            f"While parsing {values} (from {string!r}): {e}"
                        )
                    time: Optional[datetime] = None
                    match timetokens:
                        case [Time(time=time_)]:
                            time = time_
                        case [TimeFrame(start=start, end=end)]:
                            match field_:
                                case "start":
                                    time = start or end
                                case "end":
                                    time = end or start
                    if not time:
                        raise ValueError(
                            f"{values} (from {string!r}) doesn't parse to an interpretable time "
                            f"(but to {timetokens} and ignored elements {othertokens})"
                        )
                    match field_:
                        case "start":
                            return TimeStart(time=time, string=string)
                        case "end":
                            return TimeEnd(time=time, string=string)
                case (
                    "tags"
                ):  # git annex uses field 'tag' for tags, for convenience adjust it here
                    field = "tag"
                # aliases for the location field
                case str() as f if "location".startswith(
                    f
                ) or f in "at in where".split():
                    field = "location"
            valueset = frozenset(values)
            match operator:
                case "+":
                    return AddToField(field=field, values=valueset, string=string)
                case "-":
                    return RemoveFromField(field=field, values=valueset, string=string)
                case _ if values:
                    return SetField(field=field, values=valueset, string=string)
                case _:
                    return UnsetField(field=field, string=string)
        return None


@dataclass(frozen=True)
class FieldValueModifier(FieldModifier):
    values: FrozenSet[str]

    def __post_init__(self):
        object.__setattr__(self, "values", frozenset(self.values))

    @property
    def separator(self) -> Union[str, None]:
        for sep in self.SEPARATORS:
            if not any(sep in v for v in self.values):
                return sep
        return None

    @property
    def values_joined(self) -> Tuple[str, str]:
        if sep := self.separator:
            return sep, sep.join(map(str.strip, self.values))
        else:
            it = iter(self.SEPARATORS)
            sep, repl = (next(it, "") for i in range(2))
            logger.warning(
                f"Don't know what separator to use for the values in {self!r}. "
                f"None of {self.SEPARATORS!r} is safe to use they're all present in the values and we don't have an escaping mechanism. "
                f"Falling back to {sep!r} and replacing all its occurrences with {repl!r}."
            )
            return sep, sep.join(v.replace(sep, repl).strip() for v in self.values)


@dataclass(frozen=True)
class UnsetField(FieldModifier):
    def __str__(self) -> str:
        return f"{self.field}="


@dataclass(frozen=True)
class SetField(FieldValueModifier):
    def __str__(self) -> str:
        sep, joined = self.values_joined
        return f"{self.field}{sep if sep != ',' else ''}={joined}"


@dataclass(frozen=True)
class AddToField(FieldValueModifier):
    def __str__(self):
        if (
            len(self.values) == 1
            and self.field.lower() in ["tag", "tags"]
            and isinstance(Token.from_str(value := next(iter(self.values))), AddToField)
        ):
            # shortcut for tags that are not interpreted as another token
            return f"{value}"
        else:
            sep, joined = self.values_joined
            return f"{self.field}+{sep if sep != ',' else ''}={joined}"


@dataclass(frozen=True)
class RemoveFromField(FieldValueModifier):
    def __str__(self) -> str:
        sep, joined = self.values_joined
        return f"{self.field}-{sep if sep != ',' else ''}={joined}"
